import pytest

from uwsgiconf import Section
from uwsgiconf.exceptions import ConfigurationError


def test_networking_basics(assert_lines):

    assert_lines([
        'listen = 3',
    ], Section().networking.set_basic_params(queue_size=3))

    assert_lines([
        'so-keepalive = true',
    ], Section().networking.set_socket_params(keep_alive=True))

    assert_lines([
        'abstract-socket = true',
    ], Section().networking.set_unix_socket_params(abstract=True))

    assert_lines([
        'reuse-port = true',
    ], Section().networking.set_bsd_socket_params(port_reuse=True))

    assert_lines([
        'http-socket = :8080',
        'map-socket = 0:1',
    ], Section().networking.register_socket(
        address=':8080', type=Section.networking.socket_types.HTTP,
        bound_workers=1))

    assert_lines([
        'socket = :8000',
        'map-socket = 0:2,3',
    ], Section().networking.register_socket(address=':8000', bound_workers=[2, 3]))

    with pytest.raises(ConfigurationError) as einfo:
        assert_lines([
            '',
        ], Section().networking.register_socket(address='127.0.0.1', type='dummy'))
    assert 'Unknown' in str(einfo.value)  # unknown socket type

    # batch socket registration
    assert_lines([
        'socket = :8001',
        'socket = :8002',
        'socket = :8003',

    ], Section().networking.register_sockets(
        dict(address=':8001'),
        dict(address=':8002'),
        dict(address=':8003'),
    ))



from .alarms import AlarmType
from ..base import OptionsGroup, ParametrizedValue
from ..utils import make_key_val_string, listify, filter_locals
from ..exceptions import ConfigurationError


class Metric(ParametrizedValue):

    type_str = None
    type_id = None
    name_separator = ','

    def __init__(
            self, name, oid=None, alias_for=None, collector=None,
            initial_value=None, collect_interval=None, reset_after_push=None):
        """

        :param str|unicode name: Metric name.

            .. note:: Only numbers, letters, underscores, dashes and dots.

        :param str|unicode alias_for: If set metric will be a simple alias for the specified one.

        :param str|unicode oid: Metric OID.

            Required for SNMP.

            * http://uwsgi-docs.readthedocs.io/en/latest/Metrics.html#oid-assigment-for-plugins

        :param Collector collector:

        :param int initial_value: Set the metric to a specific value on startup.

        :param int collect_interval:

        :param bool reset_after_push: Reset the metric to zero (or the configured initial_value)
            after it's been pushed.

        """
        self.name = name

        if alias_for:
            # Set any type to ``alias``.
            self.type_str = MetricTypeAlias.type_str
            self.type_id = MetricTypeAlias.type_id

            if isinstance(alias_for, Metric):
                alias_for = alias_for.name

        if self.type_id == MetricTypeAlias.type_id and not alias_for:
            raise ConfigurationError('Parameter ``alias_for`` is required for ``MetricTypeAlias``.')

        type_str = self.type_str

        value = make_key_val_string(
            filter_locals(locals(), drop=['name']),
            aliases={'collect_interval': 'freq', 'type_str': 'type', 'alias_for': 'alias'}
        )

        super(Metric, self).__init__(value)

    def __str__(self):
        return 'name=%s' % super(Metric, self).__str__()


class MetricTypeCounter(Metric):
    """A generally-growing up number.

        Example:
            * number of requests

    """
    type_str = 'counter'
    type_id = 0


class MetricTypeGauge(Metric):
    """A number that can increase or decrease dynamically.

        Example:
            * memory used by a worker
            * CPU load

    """
    type_str = 'gauge'
    type_id = 1


class MetricTypeAbsolute(Metric):
    """An absolute number.

        Example:
            * memory of the whole server
            * size of the hard disk.

    """
    type_str = 'absolute'
    type_id = 2


class MetricTypeAlias(Metric):
    """This is a virtual metric pointing to another one .

    You can use it to give different names to already existing metrics.

    """
    type_str = 'alias'
    type_id = 3


class Pusher(ParametrizedValue):

    args_joiner = ','


class PusherSocket(Pusher):
    """Push metrics to a UDP server.

    Uses the following format: ``<metric> <type> <value>``
    ``<type>`` - is in the numeric form of metric type.

    """
    name = 'socket'
    plugin = 'stats_pusher_socket'

    def __init__(self, address, prefix=None):
        """
        :param str|unicode address:

        :param str|unicode prefix: Arbitrary prefix to differentiate sender.
        """
        super(PusherSocket, self).__init__(address, prefix)


class PusherRrdtool(Pusher):
    """This will store an rrd file for each metric in the specified directory.

    Each rrd file has a single data source named "metric".

    """
    name = 'rrdtool'
    plugin = 'rrdtool'

    def __init__(self, target_dir):
        """
        :param str|unicode target_dir: Directory to store rrd files into.
        """
        super(PusherRrdtool, self).__init__(target_dir)

    def set_basic_params(self, library=None, push_interval=None):
        """
        :param str|unicode library: Set the name of rrd library. Default: librrd.so.

        :param int push_interval: Set push frequency.
        """
        self._set('rrdtool-freq', push_interval)
        self._set('rrdtool-lib', library)

        return self


class PusherStatsd(Pusher):
    """Push metrics to a statsd server."""

    name = 'statsd'
    plugin = 'stats_pusher_statsd'

    def __init__(self, address, prefix=None):
        """
        :param str|unicode address:

        :param str|unicode prefix: Arbitrary prefix to differentiate sender.
        """
        super(PusherStatsd, self).__init__(address, prefix)

    def set_basic_params(self, no_workers=None, all_gauges=None):
        """
        :param bool no_workers: Disable generation of single worker metrics.

        :param bool all_gauges: Push all metrics to statsd as gauges.
        """
        self._set('statsd-no-workers', no_workers, cast=bool)
        self._set('statsd-all-gauges', all_gauges, cast=bool)

        return self


class PusherCarbon(Pusher):
    """Push metrics to a Carbon server.

    * http://uwsgi.readthedocs.io/en/latest/Carbon.html
    * http://uwsgi.readthedocs.io/en/latest/tutorials/GraphiteAndMetrics.html

    """
    name = 'carbon'
    plugin = 'carbon'

    def __init__(self, address):
        """
        :param str|unicode address:
        """
        super(PusherCarbon, self).__init__(address)

    def set_basic_params(
            self, id=None, root_node=None, push_interval=None, idle_avg_source=None,
            use_metrics=None, no_workers=None):
        """

        :param str|unicode id: Set carbon id.

        :param str|unicode root_node: Set carbon metrics root node. Default: uwsgi.

        :param int push_interval: Set carbon push frequency in seconds. Default: 60.

        :param bool no_workers: Disable generation of single worker metrics.

        :param str|unicode idle_avg_source: Average values source during idle period (no requests).
            Default: last.

            Variants:
                * last
                * zero
                * none

        :param bool use_metrics: Don't compute all statistics, use metrics subsystem data
            instead.

            .. warning:: Key names will be different.

        """
        self._set('carbon-id', id)
        self._set('carbon-root', root_node)
        self._set('carbon-freq', push_interval)
        self._set('carbon-idle-avg', idle_avg_source)
        self._set('carbon-use-metrics', use_metrics, cast=bool)
        self._set('carbon-no-workers', no_workers, cast=bool)

        return self

    def set_connection_params(
            self, timeout=None, retries=None, retries_delay=None,
            hostname_dots_replacer=None, hostname_as_address=None):
        """Sets connection related parameters.

        :param int timeout: Set carbon connection timeout in seconds. Default: 3.

        :param int retries: Set maximum number of retries in case of connection errors. Default: 1.

        :param int retries_delay: Set connection retry delay in seconds. Default: 7.

        :param str|unicode hostname_dots_replacer: Set char to use as a replacement for
            dots in hostname (dots are not replaced by default).

        :param bool hostname_as_address: Allow using hostname as carbon server address. Default: disabled.

        """
        self._set('carbon-timeout', timeout)
        self._set('carbon-max-retry', retries)
        self._set('carbon-retry-delay', retries_delay)
        self._set('carbon-hostname-dots', hostname_dots_replacer)
        self._set('carbon-name-resolve', hostname_as_address, cast=bool)

        return self


class PusherZabbix(Pusher):
    """Push metrics to a zabbix server."""

    name = 'zabbix'
    plugin = 'zabbix'

    def __init__(self, address, prefix=None):
        """
        :param str|unicode address:

        :param str|unicode prefix: Arbitrary prefix to differentiate sender.
        """
        super(PusherZabbix, self).__init__(address, prefix)

    def set_basic_params(self, template=None):
        """
        :param str|unicode template: Print (or store to a file) the zabbix template
            for the current metrics setup.
        """
        self._set('zabbix-template', template)

        return self


class PusherMongo(Pusher):
    """Push statistics (as JSON) the the specified MongoDB database."""

    name = 'mongodb'
    plugin = 'stats_pusher_mongodb'

    def __init__(self, address=None, collection=None, push_interval=None):
        """
        :param str|unicode address: Default: 127.0.0.1:27017

        :param str|unicode collection: MongoDB colection to write into. Default: uwsgi.statistics

        :param int push_interval: Write interval in seconds.
        """
        value = make_key_val_string(locals(), aliases={'push_interval': 'freq'})

        super(PusherMongo, self).__init__(value)


class PusherFile(Pusher):
    """Stores stats JSON into a file.

    .. note:: Mainly for demonstration purposes.

    """
    name = 'file'
    plugin = 'stats_pusher_file'

    def __init__(self, fpath=None, separator=None, push_interval=None):
        """
        :param str|unicode fpath: File path. Default: uwsgi.stats

        :param str|unicode separator: New entry separator. Default: \n\n

        :param int push_interval: Write interval in seconds.
        """
        value = make_key_val_string(locals(), aliases={'fpath': 'path', 'push_interval': 'freq'})

        super(PusherFile, self).__init__(value)


class Collector(ParametrizedValue):

    name_separator = ','
    name_separator_strip = True
    args_joiner = ','

    def __init__(self, *args, **kwargs):
        args = list(args)

        # Adding children for metric to use.
        children = kwargs.pop('children', None)

        if children:
            children_ = []

            for child in children:
                if isinstance(child, Metric):
                    child = child.name
                children_.append(child)

            args.append('children=%s' % ';'.join(children_))

        super(Collector, self).__init__(*args)


class CollectorManual(Collector):
    """The value must be updated manually from applications using the metrics API."""

    name = 'manual'


class CollectorPointer(Collector):
    """The value is collected from memory pointer."""

    name = 'ptr'


class CollectorFile(Collector):
    """The value is collected from a file."""

    name = 'file'

    def __init__(self, fpath, get_slot=None):
        """
        :param str|unicode fpath: File path.

        :param int get_slot: Get value from the given slot number.
            Slots: the content is split (using \n, \t, spaces, \r and zero as separator)
            and the item (the returned array is zero-based) used as the return value.

        """
        value = make_key_val_string(locals(), aliases={'fpath': 'arg1', 'get_slot': 'arg1n'})

        super(CollectorFile, self).__init__(value)


class CollectorFunction(Collector):
    """The value is computed calling a specific C function every time.

    .. note::
        * The argument it takes is a ``uwsgi_metric`` pointer.
          You generally do not need to parse the metric, so just casting to void will avoid headaches.

        * The function must returns an ``int64_t`` value.

    """

    name = 'func'

    def __init__(self, func):
        """
        :param str|unicode func: Function to call.
        """
        value = make_key_val_string(locals(), aliases={'func': 'arg1'})

        super(CollectorFunction, self).__init__(value)


class CollectorSum(Collector):
    """The value is the sum of other metrics."""

    name = 'sum'

    def __init__(self, what):
        super(CollectorSum, self).__init__(children=what)


class CollectorAvg(Collector):
    """The value is the algebraic average of the children.

    .. note:: Since 1.9.20

    """
    name = 'avg'

    def __init__(self, what):
        super(CollectorAvg, self).__init__(children=what)


class CollectorAccumulator(Collector):
    """Always add the sum of children to the final value.

    Example:

        * Round 1: child1 = 22, child2 = 17 -> metric_value = 39
        * Round 2: child1 = 26, child2 = 30 -> metric_value += 56

    """
    name = 'accumulator'

    def __init__(self, what):
        super(CollectorAccumulator, self).__init__(children=what)


class CollectorAdder(Collector):
    """Add the specified argument (arg1n) to the sum of children.

    """
    name = 'adder'

    def __init__(self, what, value):
        """
        :param int value: Value to add (multiply if it is CollectorMultiplier).
        """
        value = make_key_val_string(filter_locals(locals(), drop=['what']), aliases={'value': 'arg1n'})

        super(CollectorAdder, self).__init__(value, children=what)


class CollectorMultiplier(CollectorAdder):
    """Multiply the sum of children by the specified argument.

    Example:

        * child1 = 22, child2 = 17, arg1n = 3 -> metric_value = (22+17)*3

    """
    name = 'multiplier'


class Monitoring(OptionsGroup):
    """Monitoring facilities.

    * SNMP - http://uwsgi.readthedocs.io/en/latest/SNMP.html
    * Stats - http://uwsgi.readthedocs.io/en/latest/StatsServer.html
    * Metrics - http://uwsgi.readthedocs.io/en/latest/Metrics.html

    """

    class metric_types(object):
        """Various metric types to represent data of various nature.

        User metrics must inherit from one of those.

        """
        absolute = MetricTypeAbsolute
        alias = MetricTypeAlias
        counter = MetricTypeCounter
        gauge = MetricTypeGauge

    class collectors(object):
        """Metric collection and accumulation means."""

        accumulator = CollectorAccumulator
        adder = CollectorAdder
        avg = CollectorAvg
        file = CollectorFile
        function = CollectorFunction
        manual = CollectorManual
        multiplier = CollectorMultiplier
        pointer = CollectorPointer
        sum = CollectorSum

    class pushers(object):
        """Means to deliver metrics to various remotes or locals.

        These are available for ``.register_stats_pusher()``.

        """
        carbon = PusherCarbon
        file = PusherFile
        mongo = PusherMongo
        rrdtool = PusherRrdtool
        socket = PusherSocket
        statsd = PusherStatsd
        zabbix = PusherZabbix

    def register_metric(self, metric):
        """

        Officially Registered Metrics:

            * ``worker`` 3 - exports information about workers.
                Example: worker.1.requests **[or 3.1.1]** reports the number of requests served by worker 1.

            * ``plugin`` 4  - namespace for metrics automatically added by plugins.
                Example: plugins.foo.bar

            * ``core`` 5  - namespace for general instance information.

            * ``router`` 6 - namespace for corerouters.
                Example: router.http.active_sessions

            * ``socket`` 7 - namespace for sockets.
                Example: socket.0.listen_queue

            * ``mule`` 8 - namespace for mules.
                Example:  mule.1.signals

            * ``spooler`` 9 - namespace for spoolers.
                Example: spooler.1.signals

            * ``system`` 10 - namespace for system metrics, like loadavg or free memory.

        :param Metric|list[Metric] metric: Metric object.

        """
        for metric in listify(metric):
            self._set('metric', metric, multi=True)

        return self._section

    def set_metrics_params(self, enable=None, store_dir=None, restore=None, no_cores=None):
        """Sets basic Metrics subsystem params.

        uWSGI metrics subsystem allows you to manage "numbers" from your apps.

        When enabled, the subsystem configures a vast amount of metrics
        (like requests per-core, memory usage, etc) but, in addition to this,
        you can configure your own metrics, such as the number of active users or, say,
        hits of a particular URL, as well as the memory consumption of your app or the whole server.

        * http://uwsgi.readthedocs.io/en/latest/Metrics.html
        * SNMP Integration - http://uwsgi.readthedocs.io/en/latest/Metrics.html#snmp-integration

        :param bool enable: Enables the subsystem.

        :param str|unicode store_dir: Directory to store metrics.
            The metrics subsystem can expose all of its metrics in the form
            of text files in a directory. The content of each file is the value
            of the metric (updated in real time).

        :param bool restore: Restore previous metrics from ``store_dir``.
            When you restart a uWSGI instance, all of its metrics are reset.
            Use the option to force the metric subsystem to read-back the values
            from the metric directory before starting to collect values.

        :param bool no_cores: Disable generation of cores-related metrics.

        """
        self._set('enable-metrics', enable, cast=bool)
        self._set('metrics-dir', store_dir)
        self._set('metrics-dir-restore', restore, cast=bool)
        self._set('metrics-no-cores', no_cores, cast=bool)

        return self._section

    def set_metrics_threshold(self, name, value, check_interval=None, reset_to=None, alarm=None, alarm_message=None):
        """Sets metric threshold parameters.

        :param str|unicode name: Metric name.

        :param int value: Threshold value.

        :param int reset_to: Reset value to when threshold is reached.

        :param int check_interval: Threshold check interval in seconds.

        :param str|unicode|AlarmType alarm: Alarm to trigger when threshold is reached.

        :param str|unicode alarm_message: Message to pass to alarm. If not set metrics name is passed.

        """
        if alarm is not None and isinstance(alarm, AlarmType):
            self._section.alarms.register_alarm(alarm)
            alarm = alarm.alias

        value = make_key_val_string(
            locals(),
            aliases={
                'name': 'key',
                'reset_to': 'reset',
                'check_interval': 'rate',
                'alarm_message': 'msg',
            },
        )

        self._set('metric-threshold', value, multi=True)

        return self._section

    def set_stats_params(
            self, address=None, enable_http=None,
            minify=None, no_cores=None, no_metrics=None, push_interval=None):
        """Enables stats server on the specified address.

        * http://uwsgi.readthedocs.io/en/latest/StatsServer.html

        :param str|unicode address: Address/socket to make stats available on.

            Examples:
                * 127.0.0.1:1717
                * /tmp/statsock
                * :5050

        :param bool enable_http: Server stats over HTTP.
            Prefixes stats server json output with http headers.

        :param bool minify: Minify statistics json output.

        :param bool no_cores: Disable generation of cores-related stats.

        :param bool no_metrics: Do not include metrics in stats output.

        :param int push_interval: Set the default frequency of stats pushers in seconds/

        """
        self._set('stats-server', address)
        self._set('stats-http', enable_http, cast=bool)
        self._set('stats-minified', minify, cast=bool)
        self._set('stats-no-cores', no_cores, cast=bool)
        self._set('stats-no-metrics', no_metrics, cast=bool)
        self._set('stats-pusher-default-freq', push_interval)

        return self._section

    def register_stats_pusher(self, pusher):
        """Registers a pusher to be used for pushing statistics to various remotes/locals.

        :param Pusher|list[Pusher] pusher:

        """
        for pusher in listify(pusher):
            self._set('stats-push', pusher, multi=True)

        return self._section

    def enable_snmp(self, address, community_string):
        """Enables SNMP.

        uWSGI server embeds a tiny SNMP server that you can use to integrate
        your web apps with your monitoring infrastructure.

        * http://uwsgi.readthedocs.io/en/latest/SNMP.html

        .. note:: SNMP server is started in the master process after dropping the privileges.
            If you want it to listen on a privileged port, you can either use Capabilities on Linux,
            or use the ``as-root`` option to run the master process as root.

        :param str|unicode address: UDP address to bind to.

            Examples:

                * 192.168.1.1:2222

        :param str|unicode community_string: SNMP instance identifier to address it.

        """
        self._set('snmp', address)
        self._set('snmp-community', community_string)

        return self._section



from ..base import OptionsGroup, ParametrizedValue
from ..exceptions import ConfigurationError
from ..utils import listify, make_key_val_string, filter_locals


class Var(object):

    tpl = '%s'

    def __init__(self, name):
        self._name = name

    def __str__(self):
        return self.tpl % self._name


class RouteAction(ParametrizedValue):

    pass


class ActionFlush(RouteAction):
    """Send the current contents of the transformation buffer
    to the client (without clearing the buffer).

    * http://uwsgi.readthedocs.io/en/latest/Transformations.html#flushing-magic

    """
    name = 'flush'

    def __init__(self):
        super(ActionFlush, self).__init__()


class ActionGzip(RouteAction):
    """Encodes the response buffer to gzip."""

    name = 'gzip'
    plugin = 'transformation_gzip'

    def __init__(self):
        super(ActionGzip, self).__init__()


class ActionToFile(RouteAction):
    """Used for caching a response buffer into a static file."""

    name = 'tofile'
    plugin = 'transformation_tofile'

    def __init__(self, filename, mode=None):
        arg = make_key_val_string(locals())
        super(ActionToFile, self).__init__(arg)


class ActionUpper(RouteAction):
    """Transforms each character in uppercase.

    Mainly as an example of transformation plugin.

    """
    name = 'toupper'
    plugin = 'transformation_toupper'

    def __init__(self):
        super(ActionUpper, self).__init__()


class ActionChunked(RouteAction):
    """Encodes the output in HTTP chunked."""

    name = 'chunked'

    def __init__(self):
        super(ActionChunked, self).__init__()


class ActionTemplate(RouteAction):
    """Allows using a template file to expose everything
    from internal routing system into it.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.19.html#the-template-transformation

    """
    name = 'template'

    def __init__(self):
        super(ActionTemplate, self).__init__()


class ActionFixContentLen(RouteAction):
    """Fixes Content-length header."""

    name = 'fixcl'

    def __init__(self, add_header=False):
        """
        :param bool add_header: Force header add instead of plain fix of existing header.
        """
        if add_header:
            self.name = 'forcexcl'

        super(ActionFixContentLen, self).__init__()


class ActionDoContinue(RouteAction):
    """Stop scanning the internal routing table
    and continue to the selected request handler.

    """
    name = 'continue'

    def __init__(self):
        super(ActionDoContinue, self).__init__()


class ActionDoBreak(RouteAction):
    """Stop scanning the internal routing table and close the request."""

    name = 'break'

    def __init__(self, code, return_body=False):
        """
        :param int code: HTTP code

        :param return_body: Uses uWSGI's built-in status code and returns
            both status code and message body.

        """
        if return_body:
            self.name = 'return'

        super(ActionDoBreak, self).__init__(code)


class ActionLog(RouteAction):
    """Print the specified message in the logs or
    do not log a request is ``message`` is ``None``.

    """
    name = 'log'

    def __init__(self, message):
        """

        :param str|unicode|None message: Message to add into log.
            If ``None`` logging will be disabled for this request.

        """
        if message is None:
            self.name = 'donotlog'

        super(ActionLog, self).__init__(message)


class ActionOffloadOff(RouteAction):
    """Do not use offloading."""

    name = 'donotoffload'

    def __init__(self):
        super(ActionOffloadOff, self).__init__()


class ActionAddVarLog(RouteAction):
    """Add the specified logvar."""

    name = 'logvar'

    def __init__(self, name, val):
        """
        :param str|unicode name: Variable name.

        :param val: Variable value.
        """
        super(ActionAddVarLog, self).__init__(name, val)


class ActionDoGoto(RouteAction):
    """Make a forward jump to the specified label or rule position."""

    name = 'goto'

    def __init__(self, where):
        """
        :param str|unicode|int where: Rule number of label to go to.
        """
        super(ActionDoGoto, self).__init__(where)


class ActionAddVarCgi(RouteAction):
    """Add the specified CGI (environment) variable to the request."""

    name = 'addvar'

    def __init__(self, name, val):
        """
        :param str|unicode name: Variable name.

        :param val: Variable value.
        """
        super(ActionAddVarCgi, self).__init__(name, val)


class ActionHeaderAdd(RouteAction):
    """Add the specified HTTP header to the response."""

    name = 'addheader'

    def __init__(self, name, val):
        """
        :param str|unicode name: Header name.

        :param val: Header value.
        """
        name += ':'

        super(ActionHeaderAdd, self).__init__(name, val)


class ActionHeaderRemove(RouteAction):
    """Remove the specified HTTP header from the response."""

    name = 'delheader'

    def __init__(self, name):
        """
        :param str|unicode name: Header name.
        """
        super(ActionHeaderRemove, self).__init__(name)


class ActionHeadersOff(RouteAction):
    """Disable headers."""

    name = 'disableheaders'

    def __init__(self):
        super(ActionHeadersOff, self).__init__()


class ActionHeadersReset(RouteAction):
    """Clear the response headers, setting a new HTTP status code,
    useful for resetting a response.

    """
    name = 'clearheaders'

    def __init__(self, code):
        """
        :param int code: HTTP code.
        """
        super(ActionHeadersReset, self).__init__(code)


class ActionSignal(RouteAction):
    """Raise the specified uwsgi signal."""

    name = 'signal'

    def __init__(self, num):
        """
        :param int num: Signal number.
        """
        super(ActionSignal, self).__init__(num)


class ActionSend(RouteAction):
    """Extremely advanced (and dangerous) function allowing you
    to add raw data to the response.

    """
    name = 'send'

    def __init__(self, data, crnl=False):
        """
        :param data: Data to add to response.
        :param bool crnl: Add carriage return and new line.
        """
        if crnl:
            self.name = 'send-crnl'
        super(ActionSend, self).__init__(data)


class ActionRedirect(RouteAction):
    """Return a HTTP 301/302 Redirect to the specified URL."""

    name = 'redirect-302'
    plugin = 'router_redirect'

    def __init__(self, url, permanent=False):
        """
        :param str| unicode url: URL to redirect to.
        :param bool permanent: If ``True`` use 301, otherwise 302.
        """
        if permanent:
            self.name = 'redirect-301'

        super(ActionRedirect, self).__init__(url)


class ActionRewrite(RouteAction):
    """A rewriting engine inspired by Apache mod_rewrite.

    Rebuild PATH_INFO and QUERY_STRING according to the specified rules
    before the request is dispatched to the request handler.

    """
    name = 'rewrite'
    plugin = 'router_rewrite'

    def __init__(self, rule, do_continue=False):
        """
        :param str|unicode rule: A rewrite rule.

        :param bool do_continue: Stop request processing
            and continue to the selected request handler.

        """
        if do_continue:
            self.name = 'rewrite-last'

        super(ActionRewrite, self).__init__(rule)


class ActionRouteUwsgi(RouteAction):
    """Rewrite the modifier1, modifier2 and optionally UWSGI_APPID values of a request
    or route the request to an external uwsgi server.

    """
    name = 'uwsgi'
    plugin = 'router_uwsgi'
    args_joiner = ','

    def __init__(self, external_address='', mod1='', mod2='', app=''):
        """
        :param str|unicode external_address: External uWSGI server address (host:port).
        :param str|unicode mod1: Set modifier 1.
        :param str|unicode mod2: Set modifier 2.
        :param str|unicode app: Set ``UWSGI_APPID``.
        """
        super(ActionRouteUwsgi, self).__init__(external_address, mod1, mod2, app)


class ActionRouteExternal(RouteAction):
    """Route the request to an external HTTP server."""

    name = 'http'
    plugin = 'router_http'
    args_joiner = ','

    def __init__(self, address, host_header=None):
        """
        :param str|unicode address: External HTTP address (host:port)

        :param str|unicode host_header: HOST header value.
        """
        super(ActionRouteExternal, self).__init__(address, host_header)


class ActionAlarm(RouteAction):
    """Triggers an alarm.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.6.html#the-alarm-routing-action

    """
    name = 'alarm'

    def __init__(self, name, message):
        """
        :param str|unicode name: Alarm name

        :param str|unicode message: Message to pass into alarm.
        """
        super(ActionAlarm, self).__init__(name, message)


class ActionServeStatic(RouteAction):
    """Serve a static file from the specified physical path."""

    name = 'static'
    plugin = 'router_static'

    def __init__(self, fpath):
        """
        :param str|unicode fpath: Static file path.
        """
        super(ActionServeStatic, self).__init__(fpath)


class ActionAuthBasic(RouteAction):
    """Use Basic HTTP Auth."""

    name = 'basicauth'
    plugin = 'router_basicauth'
    args_joiner = ','

    def __init__(self, realm, user=None, password=None, do_next=False):
        """
        :param str|unicode realm:

        :param str|unicode user:

        :param str|unicode password: Password or htpasswd-like file.

        :param bool do_next: Allow next rule.
        """
        if do_next:
            self.name = 'basicauth-next'

        user_password = []

        if user:
            user += ':'
            user_password.append(user)

        if password:
            user_password.append(password)

        super(ActionAuthBasic, self).__init__(realm, ''.join(user_password) if user_password else None)


class AuthLdap(RouteAction):
    """Use Basic HTTP Auth."""

    name = 'ldapauth'
    plugin = 'ldap'
    args_joiner = ','

    def __init__(
            self, realm, address, base_dn=None, bind_dn=None, bind_password=None,
            filter=None, login_attr=None, log_level=None,
            do_next=False):
        """

        :param str|unicode realm:

        :param str|unicode address: LDAP server URI

        :param str|unicode base_dn: Base DN used when searching for users.

        :param str|unicode bind_dn: DN used for binding.
            Required if the LDAP server does not allow anonymous searches.

        :param str|unicode bind_password: Password for the ``bind_dn`` user.

        :param str|unicode filter: Filter used when searching for users. Default: ``(objectClass=*)``

        :param str|unicode login_attr: LDAP attribute that holds user login. Default: ``uid``.

        :param str|unicode log_level: Log level.

            Supported values:
                * 0 - don't log any binds
                * 1 - log authentication errors,
                * 2 - log both successful and failed binds

        :param bool do_next: Allow next rule.
        """
        arg = make_key_val_string(
            filter_locals(locals(), drop=['realm', 'do_next']),
            aliases={
                'address': 'url',
                'base_dn': 'basedn',
                'bind_dn': 'binddn',
                'bind_password': 'bindpw',
                'login_attr': 'attr',
                'log_level': 'loglevel',
            },
            items_separator=';'
        )

        if do_next:
            self.name = 'ldapauth-next'

        super(AuthLdap, self).__init__(realm, arg)


class ActionSetHarakiri(RouteAction):
    """Set harakiri timeout for the current request."""

    name = 'harakiri'

    def __init__(self, timeout):
        """
        :param int timeout:
        """
        super(ActionSetHarakiri, self).__init__(timeout)


class ActionDirChange(RouteAction):
    """Changes a directory."""

    name = 'chdir'

    def __init__(self, dir):
        """
        :param str|unicode dir: Directory to change into.
        """
        super(ActionDirChange, self).__init__(dir)


class ActionSetVarUwsgiAppid(RouteAction):
    """Set UWSGI_APPID"""

    name = 'setapp'

    def __init__(self, app):
        """
        :param str|unicode app: Application ID.
        """
        super(ActionSetVarUwsgiAppid, self).__init__(app)


class ActionSetVarRemoteUser(RouteAction):
    """Set REMOTE_USER"""

    name = 'setuser'

    def __init__(self, user):
        """
        :param str|unicode user: Username.
        """
        super(ActionSetVarRemoteUser, self).__init__(user)


class ActionSetVarUwsgiHome(RouteAction):
    """Set UWSGI_HOME"""

    name = 'sethome'

    def __init__(self, dir):
        """
        :param str|unicode dir: Directory to make a new home.
        """
        super(ActionSetVarUwsgiHome, self).__init__(dir)


class ActionSetVarUwsgiScheme(RouteAction):
    """Set UWSGI_SCHEME.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.6.html#configuring-dynamic-apps-with-internal-routing

    """
    name = 'setscheme'

    def __init__(self, value):
        """
        :param str|unicode value:
        """
        super(ActionSetVarUwsgiScheme, self).__init__(value)


class ActionSetVarScriptName(RouteAction):
    """Set SCRIPT_NAME"""

    name = 'setscriptname'

    def __init__(self, name):
        """
        :param str|unicode name: Script name
        """
        super(ActionSetVarScriptName, self).__init__(name)


class ActionSetVarRequestMethod(RouteAction):
    """Set REQUEST_METHOD"""

    name = 'setmethod'

    def __init__(self, name):
        """
        :param str|unicode name: Method name.
        """
        super(ActionSetVarRequestMethod, self).__init__(name)


class ActionSetVarRequestUri(RouteAction):
    """Set REQUEST_URI"""

    name = 'seturi'

    def __init__(self, value):
        """
        :param str|unicode value: URI
        """
        super(ActionSetVarRequestUri, self).__init__(value)


class ActionSetVarRemoteAddr(RouteAction):
    """Set REMOTE_ADDR"""

    name = 'setremoteaddr'

    def __init__(self, value):
        """
        :param str|unicode value: Address.
        """
        super(ActionSetVarRemoteAddr, self).__init__(value)


class ActionSetVarPathInfo(RouteAction):
    """Set PATH_INFO"""

    name = 'setpathinfo'

    def __init__(self, value):
        """
        :param str|unicode value: New info.
        """
        super(ActionSetVarPathInfo, self).__init__(value)


class ActionSetVarDocumentRoot(RouteAction):
    """Set DOCUMENT_ROOT"""

    name = 'setdocroot'

    def __init__(self, value):
        """
        :param str|unicode value:
        """
        super(ActionSetVarDocumentRoot, self).__init__(value)


class ActionSetUwsgiProcessName(RouteAction):
    """Set uWSGI process name."""

    name = 'setprocname'

    def __init__(self, name):
        """
        :param str|unicode name: New process name.
        """
        super(ActionSetUwsgiProcessName, self).__init__(name)


class ActionFixVarPathInfo(RouteAction):
    """Fixes PATH_INFO taking into account script name.

    This action allows you to set SCRIPT_NAME in nginx without bothering
    to rewrite the PATH_INFO (something nginx cannot afford).

    * http://uwsgi.readthedocs.io/en/latest/Changelog-2.0.11.html#fixpathinfo-routing-action

    """
    name = 'fixpathinfo'

    def __init__(self):
        super(ActionFixVarPathInfo, self).__init__()


class ActionSetScriptFile(RouteAction):
    """Set script file.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.6.html#configuring-dynamic-apps-with-internal-routing

    """
    name = 'setfile'

    def __init__(self, fpath):
        """
        :param str|unicode fpath: File path.
        """
        super(ActionSetScriptFile, self).__init__(fpath)


class SubjectBuiltin(object):

    name = None

    def __init__(self, regexp):
        self.regexp = regexp


class SubjectCustom(ParametrizedValue):
    """Represents a routing subject that supports various check."""

    args_joiner = ';'

    def __init__(self, subject, negate=False):
        """
        :param Var|str|unicode subject: Handwritten subject or a Var heir representing it.

        :param bool negate: Use to negate subject for rule.
            .. note:: You can also use tilde (~) instead of this argument for nagation.

        """
        self._subject = subject
        self.negate = negate
        super(SubjectCustom, self).__init__()

    def __invert__(self):
        self.negate = True
        return self

    def _setup(self, name, arg=None):
        self.name = '%s' % name
        self.args = [self._subject, arg]
        return self

    def exists(self):
        """Check if the subject exists in the filesystem."""
        return self._setup('exists')

    def isfile(self):
        """Check if the subject is a file."""
        return self._setup('isfile')

    def isdir(self):
        """Check if the subject is a directory."""
        return self._setup('isdir')

    def islink(self):
        """Check if the subject is a link."""
        return self._setup('islink')

    def isexec(self):
        """Check if the subject is an executable file."""
        return self._setup('isexec')

    def islord(self):
        """Check if the subject is a Legion Lord."""
        return self._setup('lord')

    def contains_ipv4(self):
        """Check if the subject is ip v4."""
        return self._setup('ipv4in')

    def contains_ipv6(self):
        """Check if the subject is ip v6."""
        return self._setup('ipv6in')

    def eq(self, val):
        """Check if the subject is equal to the specified pattern."""
        return self._setup('eq', val)

    def ge(self, val):
        """Check if the subject is greater than or equal to the specified pattern."""
        return self._setup('ishigherequal', val)

    def le(self, val):
        """Check if the subject is less than or equal to the specified pattern."""
        return self._setup('islowerequal', val)

    def gt(self, val):
        """Check if the subject is greater than the specified pattern."""
        return self._setup('ishigher', val)

    def lt(self, val):
        """Check if the subject is less than the specified pattern."""
        return self._setup('islower', val)

    def startswith(self, val):
        """Check if the subject starts with the specified pattern."""
        return self._setup('startswith', val)

    def endswith(self, val):
        """Check if the subject ends with the specified pattern."""
        return self._setup('endswith', val)

    def matches(self, regexp):
        """Check if the subject matches the specified regexp."""
        return self._setup('regexp', regexp)

    def isempty(self):
        """Check if the subject is empty."""
        return self._setup('empty')

    def contains(self, val):
        """Check if the subject contains the specified pattern."""
        return self._setup('contains', val)


class RouteRule(object):
    """Represents a routing rule."""

    class vars(object):
        """Routing variables."""

        class request(Var):
            """Returns request variable. Examples: PATH_INFO, SCRIPT_NAME, REQUEST_METHOD."""
            tpl = '${%s}'

        class metric(Var):
            """Returns metric (see ``monitoring``) variable."""
            tpl = '${metric[%s]}'

        class cookie(Var):
            """Returns cookie variable"""
            tpl = '${cookie[%s]}'

        class query(Var):
            """Returns query string variable."""
            tpl = '${qs[%s]}'

        class uwsgi(Var):
            """Returns internal uWSGI information.

             Supported variables:
                * wid
                * pid
                * uuid
                * status

            """
            tpl = '${uwsgi[%s]}'

        class time(Var):
            """Returns time/date in various forms.

            Supported variables:
                * unix

            """
            tpl = '${time[%s]}'

        class httptime(Var):
            """Returns http date adding the numeric argument (if specified)
            to the current time (use empty arg for current server time).

            """
            tpl = '${httptime[%s]}'

    class var_functions(object):
        """Functions that can be applied to variables."""

        class mime(Var):
            """Returns mime type of a variable."""
            tpl = '${mime[%s]}'

        class math(Var):
            """Perform a math operation. Example: CONTENT_LENGTH+1

            Supported operations: + - * /

            .. warning:: Requires matheval support.

            """
            tpl = '${math[%s]}'

        class base64(Var):
            """Encodes the specified var in base64"""
            tpl = '${base64[%s]}'

        class hex(Var):
            """Encodes the specified var in hex."""
            tpl = '${hex[%s]}'

        class upper(Var):
            """Uppercase the specified var."""
            tpl = '${upper[%s]}'

        class lower(Var):
            """Lowercase the specified var."""
            tpl = '${lower[%s]}'

    class stages(object):
        """During the request cycle, various stages (aka chains) are processed.

        Chains can be "recursive". A recursive chain can be called multiple times
        in a request cycle.

        """

        REQUEST = ''
        """Applied before the request is passed to the plugin."""

        ERROR = 'error'
        """Applied as soon as an HTTP status code is generate. **Recursive chain**."""

        RESPONSE = 'response'
        """Applied after the last response header has been generated (just before sending the body)."""

        FINAL = 'final'
        """Applied after the response has been sent to the client."""

    class subjects(object):
        """Routing subjects. These can be request's variables or other entities.

        .. note:: Non-custom subjects (uppercased) can be pre-optimized (during startup)
            and should be used for performance reasons.

        """
        custom = SubjectCustom

        class PATH_INFO(SubjectBuiltin):
            """Default subject, maps to PATH_INFO."""

            name = ''

        class REQUEST_URI(SubjectBuiltin):
            """Checks REQUEST_URI for a value."""

            name = 'uri'

        class QUERY_STRING(SubjectBuiltin):
            """Checks QUERY_STRING for a value."""

            name = 'qs'

        class REMOTE_ADDR(SubjectBuiltin):
            """Checks REMOTE_ADDR for a value."""

            name = 'remote-addr'

        class REMOTE_USER(SubjectBuiltin):
            """Checks REMOTE_USER for a value."""

            name = 'remote-user'

        class HTTP_HOST(SubjectBuiltin):
            """Checks HTTP_HOST for a value."""

            name = 'host'

        class HTTP_REFERER(SubjectBuiltin):
            """Checks HTTP_REFERER for a value."""

            name = 'referer'

        class HTTP_USER_AGENT(SubjectBuiltin):
            """Checks HTTP_USER_AGENT for a value."""

            name = 'user-agent'

        class STATUS(SubjectBuiltin):
            """Checks HTTP response status code.

            .. warning:: Not available in the request chain.

            """
            name = 'status'

    class transforms(object):
        """A transformation is like a filter applied to the response
        generated by your application.

        Transformations can be chained (the output of a transformation will be the input of the following one)
        and can completely overwrite response headers.

        * http://uwsgi.readthedocs.io/en/latest/Transformations.html

        """
        chunked = ActionChunked
        fix_content_len = ActionFixContentLen
        flush = ActionFlush
        gzip = ActionGzip
        template = ActionTemplate
        to_file = ActionToFile
        upper = ActionUpper

        # todo Consider adding the following and some others from sources (incl. plugins):
        # xslt, cachestore, memcachedstore, redisstore, rpc, lua

    class actions(object):
        """Actions available for routing rules.

        Values returned by actions:

            * ``NEXT`` - continue to the next rule
            * ``CONTINUE`` - stop scanning the internal routing table and run the request
            * ``BREAK`` - stop scanning the internal routing table and close the request
            * ``GOTO x`` - go to rule ``x``

        """
        add_var_cgi = ActionAddVarCgi
        add_var_log = ActionAddVarLog
        alarm = ActionAlarm
        auth_basic = ActionAuthBasic
        auth_ldap = AuthLdap
        dir_change = ActionDirChange
        do_break = ActionDoBreak
        do_continue = ActionDoContinue
        do_goto = ActionDoGoto
        fix_var_path_info = ActionFixVarPathInfo
        header_add = ActionHeaderAdd
        header_remove = ActionHeaderRemove
        headers_off = ActionHeadersOff
        headers_reset = ActionHeadersReset
        log = ActionLog
        offload_off = ActionOffloadOff
        redirect = ActionRedirect
        rewrite = ActionRewrite
        route_external = ActionRouteExternal
        route_uwsgi = ActionRouteUwsgi
        send = ActionSend
        serve_static = ActionServeStatic
        set_harakiri = ActionSetHarakiri
        set_script_file = ActionSetScriptFile
        set_uwsgi_process_name = ActionSetUwsgiProcessName
        set_var_document_root = ActionSetVarDocumentRoot
        set_var_path_info = ActionSetVarPathInfo
        set_var_remote_addr = ActionSetVarRemoteAddr
        set_var_remote_user = ActionSetVarRemoteUser
        set_var_request_method = ActionSetVarRequestMethod
        set_var_request_uri = ActionSetVarRequestUri
        set_var_script_name = ActionSetVarScriptName
        set_var_uwsgi_appid = ActionSetVarUwsgiAppid
        set_var_uwsgi_home = ActionSetVarUwsgiHome
        set_var_uwsgi_scheme = ActionSetVarUwsgiScheme
        signal = ActionSignal

        # todo Consider adding the following and some others from sources (incl. plugins):
        # cachestore, cacheset, memcached,
        # router_cache: cache, cache-continue, cachevar, cacheinc, cachedec, cachemul, cachediv
        # rpc,
        # rpc: call, rpcret, rpcnext, rpcraw, rpcvar,
        # access, spnego, radius
        # xslt, ssi, gridfs
        # cgi: cgi, cgihelper
        # router_access: access,
        # proxyhttp -router_http, proxyuwsgi -router_uwsgi, xattr -xattr
        # router_memcached: memcached, memcached-continue, memcachedstore
        # router_redis: redis, redis-continue, redisstore

    def __init__(self, action, subject=subjects.PATH_INFO('(.*)'), stage=stages.REQUEST):
        """
        :param RouteAction action: Action (or transformation) to perfrom.
            See ``.actions`` and ``.transforms``.

        :param SubjectCustom|SubjectBuiltin subject: Subject to verify before action is performed.
            See ``.subjects``.

        :param str|unicode stage: Stage on which the action needs to be performed.
            See ``.stages``.

        """
        if subject is None:
            subject = 'run'  # always run the specified route action

        subject_rule = ''

        self._custom_subject = isinstance(subject, SubjectCustom)

        if self._custom_subject:
            subject_rule = subject
            subject = 'if-not' if subject.negate else 'if'

        elif isinstance(subject, SubjectBuiltin):
            subject_rule = subject.regexp
            subject = subject.name

        self.command_label = ('%s-route-label' % stage).strip('-')
        self.command = ('%s-route-%s' % (stage, subject)).strip('-')
        self.value = subject_rule, action


class Routing(OptionsGroup):
    """Routing subsystem.

    You can use the internal routing subsystem to dynamically alter the way requests are handled.

    .. note:: Since 1.9

    * http://uwsgi.readthedocs.io/en/latest/InternalRouting.html
    * http://uwsgi.readthedocs.io/en/latest/Transformations.html

    """

    route_rule = RouteRule

    def register_route(self, route_rules, label=None):
        """Registers a routing rule.

        :param RouteRule|list[RouteRule] route_rules:

        :param str|unicode label: Label to mark the given set of rules.
            This can be used in conjunction with ``do_goto`` rule action.

            * http://uwsgi.readthedocs.io/en/latest/InternalRouting.html#goto

        """
        route_rules = listify(route_rules)

        if route_rules and label:
            self._set(route_rules[0].command_label, label, multi=True)

        for route_rules in route_rules:
            self._set(route_rules.command, route_rules.value, multi=True)

        return self._section

    def print_routes(self):
        """Print out defined routes."""

        self._set('routers-list', True, cast=bool)

        return self._section

    def set_error_page(self, status, html_fpath):
        """Add an error page (html) for managed 403, 404, 500 response.

        :param int status: HTTP status code.

        :param str|unicode html_fpath: HTML page file path.

        """
        statuses = [403, 404, 500]

        status = int(status)

        if status not in statuses:
            raise ConfigurationError(
                'Code `%s` for `routing.set_error_page()` is unsupported. Supported: %s' %
                (status, ', '.join(map(str, statuses))))

        self._set('error-page-%s' % status, html_fpath, multi=True)

        return self._section



from ..config import Section as _Section


class Section(_Section):
    """Basic nice configuration."""

    def __init__(self, name=None, touch_reload=None, workers=None, threads=None, mules=None, **kwargs):
        """

        :param str|unicode name: Section name.

        :param str|list touch_reload: Reload uWSGI if the specified file or directory is modified/touched.

        :param int workers: Spawn the specified number of workers (processes).
            Default: workers number equals to CPU count.

        :param int threads: Number of threads per worker.

        :param int mules: Number of mules to spawn.

        :param kwargs:
        """
        super(Section, self).__init__(strict_config=True, name=name, **kwargs)

        if touch_reload:
            self.main_process.set_basic_params(touch_reload=touch_reload)

        if workers:
            self.workers.set_basic_params(count=workers)
        else:
            self.workers.set_count_auto()

        self.workers.set_thread_params(count=threads)
        self.workers.set_mules_params(mules=mules)
        self.main_process.set_basic_params(vacuum=True)
        self.main_process.set_naming_params(autonaming=True)
        self.master_process.set_basic_params(enable=True)
        self.locks.set_basic_params(thunder_lock=True)


class PythonSection(Section):
    """Basic nice configuration using Python plugin."""

    def __init__(self, name=None, basic_params_python=None, wsgi_module=None, **kwargs):
        """

        :param str|unicode name: Section name.

        :param dict basic_params_python: See Python plugin params.

        :param str|unicode wsgi_module: wsgi application module path or filepath

            Example:
                mypackage.my_wsgi_module -- read from `application` attr of mypackage/my_wsgi_module.py
                mypackage.my_wsgi_module:my_app -- read from `my_app` attr of mypackage/my_wsgi_module.py

        :param kwargs:
        """
        super(PythonSection, self).__init__(name=name, basic_params_python=basic_params_python, **kwargs)

        if wsgi_module:
            self.python.set_wsgi_params(module=wsgi_module)



from ..base import OptionsGroup


class Queue(OptionsGroup):
    """Queue.

    At the low level it is a simple block-based shared array,
    with two optional counters, one for stack-style, LIFO usage,
    the other one for FIFO.

    http://uwsgi-docs.readthedocs.io/en/latest/Queue.html

    """

    def enable(self, size, block_size=None, store=None, store_sync_interval=None):
        """Enables shared queue of the given size.

        :param int size: Queue size.

        :param int block_size: Block size in bytes. Default: 8 KiB.

        :param str|unicode store: Persist the queue into file.

        :param int store_sync_interval: Store sync interval in master cycles (usually seconds).

        """
        self._set('queue', size)
        self._set('queue-blocksize', block_size)
        self._set('queue-store', store)
        self._set('queue-store-sync', store_sync_interval)

        return self._section



import pytest


@pytest.fixture(scope='session')
def ini():

    def wrapped(section, stamp):
        return section.as_configuration().format(stamp=stamp)

    return wrapped


@pytest.fixture(scope='session')
def assert_lines(ini):

    def wrapped(lines, source, assert_in=True, stamp=False):
        source = ini(source, stamp)

        if not isinstance(lines, list):
            lines = [lines]

        for line in lines:
            if assert_in:
                assert line in source
            else:
                assert line not in source

    return wrapped



from ..base import OptionsGroup
from ..exceptions import ConfigurationError
from ..utils import listify


class Networking(OptionsGroup):
    """Networking related stuff. Socket definition, binding and tuning."""
    
    class socket_types(object):
        """Available socket types to use with ``.register_socket()``."""
    
        DEFAULT = 'default'
        """Bind using default protocol. See ``default_protocol`` option."""

        UWSGI = 'uwsgi'

        HTTP = 'http'
        """Bind to the specified socket using HTTP"""
    
        HTTP11 = 'http11'  # Keep-Alive
    
        UDP = 'udp'
        """Run the udp server on the specified address.
        
        .. note:: Mainly useful for SNMP or shared UDP logging.
        
        """
    
        FASTCGI = 'fastcgi'
        """Bind to the specified socket using FastCGI."""
    
        SCGI = 'scgi'
        """Bind to the specified UNIX/TCP socket using SCGI protocol."""
    
        RAW = 'raw'
        """Bind to the specified UNIX/TCP socket using RAW protocol."""
    
        SHARED = 'shared'
        """Create a shared socket for advanced jailing or IPC purposes.
        
        Allows you to create a socket early in the server's startup 
        and use it after privileges drop or jailing. This can be used 
        to bind to privileged (<1024) ports.
        
        """
    
        ZERO_MQ = 'zmq'
        """Introduce zeromq pub/sub pair."""

    class modes(object):
        """Available socket modes to use with ``.register_socket()``."""

        SSL = '+ssl'
        """Use SSL."""

        NPH = '+nph'
        """Bind to the specified UNIX/TCP socket using nph mode."""

        PERSISTENT = '+persistent'
        """Use persistent uwsgi protocol (puwsgi)."""

        UNDEFERRED = '+undeferred'
        """Use shared socket undeferred mode."""

    def __init__(self, *args, **kwargs):
        super(Networking, self).__init__(*args, **kwargs)

        self._workers_binding = {}
        self._current_socket_idx = 0

    def set_basic_params(self, queue_size=None, freebind=None, default_socket_type=None):
        """

        :param int queue_size: Every socket has an associated queue where request will be put waiting
            for a process to became ready to accept them. When this queue is full, requests will be rejected.
            Default 100.

            .. note:: The maximum value is system/kernel dependent.

        :param bool freebind: Put socket in freebind mode.
            Allows binding to non-existent network addresses.

            .. note:: Linux only.

        :param str|unicode default_socket_type: Force the socket type as default.
            See ``.socket_types``.

        """
        self._set('listen', queue_size)
        self._set('freebind', freebind, cast=bool)
        self._set('socket-protocol', default_socket_type)

        return self._section

    def set_socket_params(
            self, send_timeout=None, keep_alive=None, no_defer_accept=None,
            buffer_send=None, buffer_receive=None):
        """Sets common socket params.

        :param int send_timeout: Send (write) timeout in seconds.

        :param bool keep_alive: Enable TCP KEEPALIVEs.

        :param bool no_defer_accept: Disable deferred ``accept()`` on sockets
            by default (where available) uWSGI will defer the accept() of requests until some data
            is sent by the client (this is a security/performance measure).
            If you want to disable this feature for some reason, specify this option.

        :param int buffer_send: Set SO_SNDBUF (bytes).

        :param int buffer_receive: Set SO_RCVBUF (bytes).

        """
        self._set('so-send-timeout', send_timeout)
        self._set('so-keepalive', keep_alive, cast=bool)
        self._set('no-defer-accept', no_defer_accept, cast=bool)
        self._set('socket-sndbuf', buffer_send)
        self._set('socket-rcvbuf', buffer_receive)

        return self._section

    def set_unix_socket_params(self, abstract=None, permissions=None, owner=None, umask=None):
        """Sets Unix-socket related params.

        :param bool abstract: Force UNIX socket into abstract mode (Linux only).

        :param str permissions: UNIX sockets are filesystem objects that obey
            UNIX permissions like any other filesystem object.

            You can set the UNIX sockets' permissions with this option if your webserver
            would otherwise have no access to the uWSGI socket. When used without a parameter,
            the permissions will be set to 666. Otherwise the specified chmod value will be used.

        :param str owner: Chown UNIX sockets.

        :param str umask: Set UNIX socket umask.

        """
        self._set('abstract-socket', abstract, cast=bool)
        self._set('chmod-socket', permissions)
        self._set('chown-socket', owner)
        self._set('umask', umask)

        return self._section

    def set_bsd_socket_params(self, port_reuse=None):
        """Sets BSD-sockets related params.

        :param bool port_reuse: Enable REUSE_PORT flag on socket to allow multiple
            instances binding on the same address (BSD only).

        """
        self._set('reuse-port', port_reuse, cast=bool)

        return self._section

    def register_sockets(self, *reg_dicts):
        """Convenience-method for batch socket registration.

        :param reg_dicts: Registration dictionaries with the same keys,
            which are accepted by .register_socket().

        """
        for reg_dict in reg_dicts:
            self.register_socket(**reg_dict)

        return self._section

    def register_socket(self, address, type=socket_types.DEFAULT, mode=None, bound_workers=None):
        """Registers a socket.

        :param str address: Address to bind socket to.
            Examples:
                * socket file - /tmp/uwsgi.sock
                * all interfaces - 0.0.0.0:8080
                * all interfaces - :9090
                * ssl files - :9090,foobar.crt,foobar.key

        :param str type: Socket type. See Networking.sockets

        :param str mode: Socket mode. See Networking.modes

        :param str|int|list bound_workers: Map socket to specific workers.
            As you can bind a uWSGI instance to multiple sockets, you can use this option to map
            specific workers to specific sockets to implement a sort of in-process Quality of Service scheme.
            If you host multiple apps in the same uWSGI instance, you can easily dedicate resources to each of them.

        """
        # todo maybe a convenience method for ssl - sertificate + key
        # todo *-modifier1

        mode = mode or ''

        param_name = {
            self.socket_types.UWSGI: {
                self.modes.SSL: 'suwsgi-socket',
                self.modes.PERSISTENT: 'puwsgi-socket',

            }.get(mode, 'uwsgi-socket'),
            # Default: Bind to the specified socket with default protocol (see `protocol/socket-protocol`)
            # socket-protocol = 0,uwsgi
            # socket-protocol = 3,uwsgidump

            self.socket_types.HTTP: 'https-socket' if mode == self.modes.SSL else 'http-socket',

            self.socket_types.HTTP11: 'http11-socket',

            self.socket_types.FASTCGI: 'fastcgi-nph-socket' if mode == self.modes.NPH else 'fastcgi-socket',

            self.socket_types.SCGI: 'scgi-nph-socket' if mode == self.modes.NPH else 'scgi-socket',

            self.socket_types.RAW: 'raw-socket',

            self.socket_types.SHARED: 'undeferred-shared-socket' if mode == self.modes.UNDEFERRED else 'shared-socket',

            self.socket_types.UDP: 'upd',

            self.socket_types.ZERO_MQ: 'zeromq-socket',

            self.socket_types.DEFAULT: 'socket',

        }.get(type)

        if param_name is None:
            raise ConfigurationError('Unknown socket type: `%s`' % type)

        self._set(param_name, address, multi=True)

        if bound_workers:
            bound_workers = listify(bound_workers)

            self._set(
                'map-socket', '%s:%s' % (self._current_socket_idx, ','.join(map(str, bound_workers))),
                multi=True)

            self._current_socket_idx += 1

        return self._section



from uwsgiconf import Section
from uwsgiconf.exceptions import ConfigurationError

import pytest


def test_monitoring_snmp(assert_lines):

    assert_lines([
        'snmp = 192.168.1.1:2222',
        'snmp-community = mymetr',
    ], Section().monitoring.enable_snmp('192.168.1.1:2222', 'mymetr'))


def test_monitoring_stats(assert_lines):

    assert_lines([
        'stats-server = 127.0.0.1:1717',
        'stats-http = true',
    ], Section().monitoring.set_stats_params('127.0.0.1:1717', enable_http=True))


def test_monitoring_metrics(assert_lines):

    assert_lines([
        'enable-metrics = true',
        'metrics-dir = /here',
    ], Section().monitoring.set_metrics_params(enable=True, store_dir='/here'))

    assert_lines([
        'metric-threshold = alarm=some,key=mycounter,reset=0,value=1000',
    ], Section().monitoring.set_metrics_threshold('mycounter', 1000, reset_to=0, alarm='some'))

    section = Section()
    assert_lines([
        'alarm = pinger cmd:ping 127.0.0.1',
        'metric-threshold = alarm=pinger,key=mycounter,value=2000',
    ], section.monitoring.set_metrics_threshold(
        'mycounter', 2000, alarm=section.alarms.alarm_types.command('pinger', 'ping 127.0.0.1')
    ))


def test_monitoring_register_metric(assert_lines):

    monitoring = Section().monitoring

    my_abs = monitoring.metric_types.absolute('myabs', initial_value=100)
    my_abs2 = monitoring.metric_types.alias('myabs2', initial_value=200, alias_for=my_abs)

    with pytest.raises(ConfigurationError):
        # Alias without alias_for
        monitoring.metric_types.alias('myabs3', initial_value=200)

    assert_lines([
        'metric = name=myabs,initial_value=100,type=absolute',
        'metric = name=myabs2,alias=myabs,initial_value=200,type=alias',
        'metric = name=mymet,collector=sum,children=worker.1.requests;myabs,type=gauge',

    ], monitoring.register_metric([
        my_abs,
        my_abs2,
        monitoring.metric_types.gauge(
            'mymet',
            collector=monitoring.collectors.sum(['worker.1.requests', my_abs]),
        )
    ]))


def test_monitoring_collectors(assert_lines):

    collectors = Section().monitoring.collectors

    assert str(collectors.manual()) == 'manual'
    assert str(collectors.pointer()) == 'ptr'
    assert str(collectors.file('/here/a', get_slot=2)) == 'file,arg1=/here/a,arg1n=2'
    assert str(collectors.function('some')) == 'func,arg1=some'
    assert str(collectors.sum(['a', 'b'])) == 'sum,children=a;b'
    assert str(collectors.avg(['a', 'b'])) == 'avg,children=a;b'
    assert str(collectors.accumulator(['a', 'b'])) == 'accumulator,children=a;b'
    assert str(collectors.adder(['a', 'b'], 3)) == 'adder,arg1n=3,children=a;b'
    assert str(collectors.multiplier(['a', 'b'], 4)) == 'multiplier,arg1n=4,children=a;b'


def test_monitoring_pushers(assert_lines):

    monitoring = Section().monitoring
    assert_lines([
        'stats-push = socket:127.0.0.1:8125,myinstance',

    ], monitoring.register_stats_pusher(
        monitoring.pushers.socket('127.0.0.1:8125', 'myinstance')
    ))

    monitoring = Section().monitoring
    pusher = monitoring.pushers.rrdtool('/here')
    pusher.set_basic_params('libmyrdd.so')
    assert_lines([
        'stats-push = rrdtool:/here',
        'rrdtool-lib = libmyrdd.so',

    ], monitoring.register_stats_pusher(pusher))

    monitoring = Section().monitoring
    pusher = monitoring.pushers.statsd('127.0.0.1:8125', 'myinstance')
    pusher.set_basic_params(no_workers=True)
    assert_lines([
        'stats-push = statsd:127.0.0.1:8125,myinstance',
        'statsd-no-workers = true',

    ], monitoring.register_stats_pusher(
        pusher
    ))

    monitoring = Section().monitoring
    pusher = monitoring.pushers.carbon('127.0.0.1:2003')
    pusher.set_basic_params(root_node='myroot')
    pusher.set_connection_params(retries=5)
    assert_lines([
        'stats-push = carbon:127.0.0.1:2003',

    ], monitoring.register_stats_pusher(
        pusher
    ))

    monitoring = Section().monitoring
    pusher = monitoring.pushers.zabbix('127.0.0.1:10051')
    pusher.set_basic_params('/put/here')
    assert_lines([
        'stats-push = zabbix:127.0.0.1:10051',
        'zabbix-template = /put/here',

    ], monitoring.register_stats_pusher(
        pusher
    ))

    monitoring = Section().monitoring
    assert_lines([
        'stats-push = mongodb:collection=mycol,freq=10',

    ], monitoring.register_stats_pusher(
        monitoring.pushers.mongo(collection='mycol', push_interval=10)
    ))

    monitoring = Section().monitoring
    assert_lines([
        'stats-push = file:path=/here/a.json,freq=15',

    ], monitoring.register_stats_pusher(
        monitoring.pushers.file('/here/a.json', push_interval=15)
    ))



from ..base import OptionsGroup


class Applications(OptionsGroup):
    """Applications.

    """

    def set_basic_params(
            self, exit_if_none=None, max_per_worker=None, single_interpreter=None, no_default=None):
        """

        :param bool exit_if_none: Exit if no app can be loaded.

        :param int max_per_worker: Set the maximum number of per-worker applications.

        :param bool single_interpreter: Do not use multiple interpreters (where available).
            Some of the supported languages (such as Python) have the concept of "multiple interpreters".
            By default every app is loaded in a new python interpreter (that means a pretty-well isolated
            namespace for each app). If you want all of the app to be loaded in the same python vm,
            use the this option.

        :param bool no_default: Do not fallback to default app.

        """

        self._set('need-app', exit_if_none, cast=bool)
        self._set('max-apps', max_per_worker)
        self._set('single-interpreter', single_interpreter, cast=bool)
        self._set('no-default-app', no_default, cast=bool)

        return self._section

    def change_dir(self, to, after_load=False):
        """Chdir to specified directory before or after apps loading.

        :param str|unicode to: Target directory.

        :param bool after_load:
                *True* - after load
                *False* - before load

        """
        self._set('chdir2' if after_load else 'chdir', to)

        return self._section

    def mount(self, mountpoint, app, into_worker=False):
        """Load application under mountpoint.

        Example:
            * .mount('/app1', 'app1.py')
            * .mount('example.com', 'app2.py')
            * .mount('the_app3', 'app3.py')
            * .mount('/pinax/here', '/var/www/pinax/deploy/pinax.wsgi')

        * http://uwsgi-docs.readthedocs.io/en/latest/Nginx.html#hosting-multiple-apps-in-the-same-process-aka-managing-script-name-and-path-info

        :param str|unicode mountpoint: Host name, URL part, etc.

        :param str|unicode app: App module/file.

        :param bool into_worker: Load application under mountpoint
            in the specified worker or after workers spawn.

        """
        # todo check worker mount -- uwsgi_init_worker_mount_app() expects worker://
        self._set('worker-mount' if into_worker else 'mount', '%s=%s' % (mountpoint, app), multi=True)

        return self._section

    def switch_into_lazy_mode(self, affect_master=None):
        """Load apps in workers instead of master.

        This option may have memory usage implications
        as Copy-on-Write semantics can not be used.

        .. note:: Consider using ``touch_chain_reload`` option in ``workers`` basic params
            for lazy apps reloading.

        :param bool affect_master: If **True** only workers will be
          reloaded by uWSGI's reload signals; the master will remain alive.

          .. warning:: uWSGI configuration changes are not picked up on reload by the master.


        """
        self._set('lazy' if affect_master else 'lazy-apps', True, cast=bool)

        return self._section



from ..base import OptionsGroup, ParametrizedValue
from ..utils import listify


class HookAction(ParametrizedValue):

    pass


class ActionMount(HookAction):
    """Mount or unmount filesystems.

    Examples:
        * Mount: proc none /proc
        * Unmount: /proc

    """

    name = 'mount'

    def __init__(self, mountpoint, fs=None, src=None, flags=None):
        """

        :param str|unicode mountpoint:

        :param str|unicode fs: Filesystem. Presence indicates mounting.

        :param str|unicode src: Presence indicates mounting.

        :param str|unicode|list flags: Flags available for the operating system.
            As an example on Linux you will options like: bind, recursive, readonly, rec, detach etc.

        """
        if flags is not None:
            flags = listify(flags)
            flags = ','.join(flags)

        if fs:
            args = [fs, src, mountpoint, flags]

        else:
            args = [mountpoint, flags]
            self.name = 'umount'

        super(ActionMount, self).__init__(*args)


class ActionExecute(HookAction):
    """Run the shell command.

    Command run under ``/bin/sh``.
    If for some reason you do not want to use ``/bin/sh``,
    use ``binsh`` option,

    Examples:
        * cat /proc/self/mounts

    """

    name = 'exec'

    # todo consider adding safeexec
    def __init__(self, command):
        super(ActionExecute, self).__init__(command)


class ActionCall(HookAction):
    """Call functions in the current process address space."""

    name = 'call'

    def __init__(self, target, honour_exit_status=False, arg_int=False):
        """
        :param str|unicode target: Symbol and args.

        :param bool honour_exit_status: Expect an int return.
            Anything != 0 means failure.

        :param bool arg_int: Parse the argument as an int.

        """
        name = self.name

        if arg_int:
            name += 'int'

        if honour_exit_status:
            name += 'ret'

        self.name = name

        super(ActionCall, self).__init__(target)


class ActionDirChange(HookAction):
    """Changes a directory.

    Convenience action, same as ``call:chdir <directory>``.

    """
    name = 'cd'

    def __init__(self, target_dir):
        super(ActionDirChange, self).__init__(target_dir)


class ActionDirCreate(HookAction):
    """Creates a directory with 0777."""

    name = 'mkdir'

    def __init__(self, target_dir):
        super(ActionDirCreate, self).__init__(target_dir)


class ActionFileCreate(HookAction):
    """Creates a directory with 0666."""

    name = 'create'

    def __init__(self, fpath):
        super(ActionFileCreate, self).__init__(fpath)


class ActionExit(HookAction):
    """Exits.

    Convenience action, same as ``callint:exit [num]``.

    """
    name = 'exit'

    def __init__(self, status_code=None):
        super(ActionExit, self).__init__(status_code)


class ActionPrintout(HookAction):
    """Prints.

    Convenience action, same as calling the ``uwsgi_log`` symbol.

    """
    name = 'print'

    def __init__(self, text=None):
        super(ActionPrintout, self).__init__(text)


class ActionSetHostName(HookAction):
    """Sets a host name."""

    name = 'hostname'

    def __init__(self, name):
        super(ActionSetHostName, self).__init__(name)


class ActionAlarm(HookAction):
    """Issues an alarm. See ``.alarms`` options group."""

    name = 'alarm'

    def __init__(self, alarm, message):
        super(ActionAlarm, self).__init__(alarm, message)


class ActionFileWrite(HookAction):
    """Writes a string to the specified file.

    If file doesn't exist it will be created.

    .. note:: Since 1.9.21

    """
    name = 'write'

    def __init__(self, target, text, append=False, newline=False):
        """

        :param str|unicode target: File to write to.

        :param str|unicode text: Text to write into file.

        :param bool append: Append text instead of rewrite.

        :param bool newline: Add a newline at the end.

        """
        if append:
            self.name = 'append'

        if newline:
            self.name += 'n'

        super(ActionFileWrite, self).__init__(target, text)


class ActionFifoWrite(HookAction):
    """Writes a string to the specified FIFO (see ``fifo_file`` from ``master_process`` params)."""

    name = 'writefifo'

    def __init__(self, target, text, wait=False):
        """
        :param bool wait: Wait until FIFO is available.

        """
        if wait:
            self.name = 'spinningfifo'

        super(ActionFifoWrite, self).__init__(target, text)


class ActionUnlink(HookAction):
    """Unlink the specified file.

    .. note:: Since 1.9.21

    """
    name = 'unlink'

    def __init__(self, target):
        super(ActionUnlink, self).__init__(target)


class MainProcess(OptionsGroup):
    """Main process is the uWSGI process.

    .. warning:: Do not run uWSGI instances as root.
        You can start your uWSGIs as root, but be sure to drop privileges
        with the ``uid`` and ``gid`` options from ``set_owner_params``.

    """

    class actions(object):
        """Actions available for ``.set_hook()``."""

        alarm = ActionAlarm
        call = ActionCall
        dir_change = ActionDirChange
        dir_create = ActionDirCreate
        execute = ActionExecute
        exit = ActionExit
        fifo_write = ActionFifoWrite
        file_create = ActionFileCreate
        file_write = ActionFileWrite
        mount = ActionMount
        printout = ActionPrintout
        set_host_name = ActionSetHostName
        unlink = ActionUnlink

        # todo consider adding:
        # putenv, chmod/sticky, chown/chown2, rpc/retryrpc, unix_signal
        # wait_for_fs/wait_for_file/wait_for_dir, wait_for_socket

    class phases:
        """Phases available for hooking using ``.set_hook()``.

        Some of them may be **fatal** - a failing hook for them
        will mean failing of the whole uWSGI instance (generally calling exit(1)).

        """

        ASAP = 'asap'
        """As soon as possible. **Fatal**
        
        Run directly after configuration file has been parsed, before anything else is done.
        
        """

        JAIL_PRE = 'pre-jail'
        """Before jailing. **Fatal** 
        
        Run before any attempt to drop privileges or put the process in some form of jail.        
        
        """

        JAIL_IN = 'in-jail'
        """In jail after initialization. **Fatal** 

        Run soon after jayling, but after post-jail. 
        If jailing requires fork(), the chidlren run this phase.
        
        """

        JAIL_POST = 'post-jail'
        """After jailing. **Fatal**
        
        Run soon after any jailing, but before privileges drop. 
        If jailing requires fork(), the parent process run this phase.
        
        """

        PRIV_DROP_PRE = 'as-root'
        """Before privileges drop. **Fatal**
        
        Last chance to run something as root.
        
        """

        PRIV_DROP_POST = 'as-user'
        """After privileges drop. **Fatal**"""

        MASTER_START = 'master-start'
        """When Master starts."""

        EMPEROR_START = 'emperor-start'
        """When Emperor starts."""

        EMPEROR_STOP = 'emperor-stop'
        """When Emperor sent a stop message."""

        EMPEROR_RELOAD = 'emperor-reload'
        """When Emperor sent a reload message."""

        EMPEROR_LOST = 'emperor-lost'
        """When Emperor connection is lost."""

        EXIT = 'as-user-atexit'
        """Before app exit and reload."""

        APP_LOAD_PRE = 'pre-app'
        """Before app loading. **Fatal**"""

        APP_LOAD_POST = 'post-app'
        """After app loading. **Fatal**"""

        VASSAL_ON_DEMAND_IN = 'as-on-demand-vassal'
        """Whenever a vassal enters on-demand mode."""

        VASSAL_CONFIG_CHANGE_POST = 'as-on-config-vassal'
        """Whenever the emperor detects a config change for an on-demand vassal."""

        VASSAL_START_PRE = 'as-emperor-before-vassal'
        """Before the new vassal is spawned."""

        VASSAL_PRIV_DRP_PRE = 'as-vassal-before-drop'
        """In vassal, before dropping its privileges."""

        VASSAL_SET_NAMESPACE = 'as-emperor-setns'
        """In the emperor entering vassal namespace."""

        VASSAL_START_IN = 'as-vassal'
        """In the vassal before executing the uwsgi binary. **Fatal** 
        
        In vassal on start just before calling exec() directly in the new namespace.
        
        """

        VASSAL_START_POST = 'as-emperor'
        """In the emperor soon after a vassal has been spawn.
         
        Setting 4 env vars:
            * UWSGI_VASSAL_CONFIG 
            * UWSGI_VASSAL_PID
            * UWSGI_VASSAL_UID 
            * UWSGI_VASSAL_GID

        """

        GATEWAY_START_IN_EACH = 'as-gateway'
        """In each gateway on start."""

        MULE_START_IN_EACH = 'as-mule'
        """In each mule on start."""

        WORKER_ACCEPTING_PRE_EACH = 'accepting'
        """Before the each worker starts accepting requests. 
        
        .. note:: Since 1.9.21
        
        """

        WORKER_ACCEPTING_PRE_FIRST = 'accepting1'
        """Before the first worker starts accepting requests.
        
        .. note:: Since 1.9.21
        
        """

        WORKER_ACCEPTING_PRE_EACH_ONCE = 'accepting-once'
        """Before the each worker starts accepting requests, one time per instance. 
        
        .. note:: Since 1.9.21
        
        """

        WORKER_ACCEPTING_PRE_FIRST_ONCE = 'accepting1-once'
        """Before the first worker starts accepting requests, one time per instance.
        
        .. note:: Since 1.9.21
        
        """

    def set_basic_params(
            self, touch_reload=None, priority=None, vacuum=None, binary_path=None, honour_stdin=None):
        """

        :param str|list touch_reload: Reload uWSGI if the specified file or directory is modified/touched.

        :param int priority: Set processes/threads priority (``nice``) value.

        :param bool vacuum: Try to remove all of the generated files/sockets
            (UNIX sockets and pidfiles) upon exit.

        :param str|unicode binary_path: Force uWSGI binary path.
            If you do not have uWSGI in the system path you can force its path with this option
            to permit the reloading system and the Emperor to easily find the binary to execute.

        :param bool honour_stdin: Do not remap stdin to ``/dev/null``.
            By default, ``stdin`` is remapped to ``/dev/null`` on uWSGI startup.
            If you need a valid stdin (for debugging, piping and so on) use this option.

        """
        self._set('touch-reload', touch_reload, multi=True)
        self._set('prio', priority)

        self._set('vacuum', vacuum, cast=bool)
        self._set('binary-path', binary_path)
        self._set('honour-stdin', honour_stdin, cast=bool)

        return self._section

    def daemonize(self, log_into, after_app_loading=False):
        """Daemonize uWSGI.

        :param str|unicode log_into: Logging destination:

            * File: /tmp/mylog.log

            * UPD: 192.168.1.2:1717

                .. note:: This will require an UDP server to manage log messages.
                    Use ``networking.register_socket('192.168.1.2:1717, type=networking.SOCK_UDP)``
                    to start uWSGI UDP server.

        :param str|unicode bool after_app_loading: Whether to daemonize after
            or before applications loading.

        """
        self._set('daemonize2' if after_app_loading else 'daemonize', log_into)

        return self._section

    def set_owner_params(self, uid=None, gid=None, add_gids=None, set_asap=False):
        """Set process owner params - user, group.

        :param str|unicode|int uid: Set uid to the specified username or uid.

        :param str|unicode|int gid: Set gid to the specified groupname or gid.

        :param list|str|unicode|int add_gids: Add the specified group id to the process credentials.
            This options allows you to add additional group ids to the current process.
            You can specify it multiple times.

        :param bool set_asap: Set as soon as possible.
            Setting them on top of your vassal file will force the instance to setuid()/setgid()
            as soon as possible and without the (theoretical) possibility to override them.

        """
        prefix = 'immediate-' if set_asap else ''

        self._set(prefix + 'uid', uid)
        self._set(prefix + 'gid', gid)
        self._set('add-gid', add_gids, multi=True)

        return self._section

    def set_hook(self, phase, action):
        """Allows setting hooks (attaching actions) for various uWSGI phases.

        :param str|unicode phase: See constants in ``.phases``.

        :param str|unicode|list|HookAction|list[HookAction] action:

        """
        self._set('hook-%s' % phase, action, multi=True)

        return self._section

    def set_hook_touch(self, fpath, action):
        """Allows running certain action when the specified file is touched.

        :param str|unicode fpath: File path.

        :param str|unicode|list|HookAction|list[HookAction] action:

        """
        self._set('hook-touch', '%s %s' % (fpath, action), multi=True)

        return self._section

    def set_hook_after_request(self, func):
        """Run the specified function/symbol (C level) after each request.

        :param str|unicode func:

        """
        self._set('after-request-hook', func, multi=True)

        return self._section

    def set_on_exit_params(self, skip_hooks=None, skip_teardown=None):
        """Set params related to process exit procedure.

        :param bool skip_hooks: Skip ``EXIT`` phase hook.

            .. note:: Ignored by the master.

        :param bool skip_teardown: Allows skipping teardown (finalization) processes for some plugins.

            .. note:: Ignored by the master.

            Supported by:
                * Perl
                * Python

        """
        self._set('skip-atexit', skip_hooks, cast=bool)
        self._set('skip-atexit-teardown', skip_teardown, cast=bool)

        return self._section

    def run_command_on_event(self, command, phase=phases.ASAP):
        """Run the given command on a given phase.

        :param str|unicode command:

        :param str|unicode phase: See constants in ``Phases`` class.

        """
        self._set('exec-%s' % phase, command, multi=True)

        return self._section

    def run_command_on_touch(self, command, target):
        """Run command when the specified file is modified/touched.

        :param str|unicode command:

        :param str|unicode target: File path.

        """
        self._set('touch-exec', '%s %s' % (target, command), multi=True)

        return self._section

    def set_pid_file(self, fpath, before_priv_drop=True, safe=False):
        """Creates pidfile before or after privileges drop.

        :param str|unicode fpath: File path.

        :param bool before_priv_drop:

        :param bool safe: The safe-pidfile works similar to pidfile
            but performs the write a little later in the loading process.
            This avoids overwriting the value when app loading fails,
            with the consequent loss of a valid PID number.

        """
        command = 'pidfile'

        if not before_priv_drop:
            command += '2'

        if safe:
            command = 'safe-' + command

        self._set(command, fpath)

        return self._section

    def set_naming_params(self, autonaming=None, prefix=None, suffix=None, name=None):
        """Setups processes naming parameters.

        :param bool autonaming: Automatically set process name to something meaningful.
            Generated process names may be 'uWSGI Master', 'uWSGI Worker #', etc.

        :param str prefix: Add prefix to process names.

        :param str suffix: Append string to process names.

        :param str name: Set process names to given static value.

        """
        self._set('auto-procname', autonaming, cast=bool)
        self._set('procname-prefix%s' % ('-spaced' if prefix and prefix.endswith(' ') else ''), prefix)
        self._set('procname-append', suffix)
        self._set('procname', name)

        return self._section



from uwsgiconf import Section


def test_logging_basics(assert_lines):

    logging = Section().logging

    assert_lines([
        'disable-logging = true',
        'log-format = %(method) --> %(uri)',

    ], logging.set_basic_params(
        no_requests=True,
        template='%s --> %s' % (logging.vars.REQ_METHOD, logging.vars.REQ_URL)
    ))

    assert_lines([
        'log-reopen = true',
    ], Section().logging.set_file_params(reopen_on_reload=True))

    assert_lines([
        'log-master = true',
        'log-master-req-stream = true',
    ], Section().logging.set_master_logging_params(
        enable=True, sock_stream_requests_only=True))

    assert_lines([
        'loggers-list = true',
    ], Section().logging.print_loggers())

    assert_lines([
        'log-route = socket aline',
    ], Section().logging.add_logger_route('socket', 'aline'))


def test_logging_set_filters(assert_lines):

    assert_lines([
        'log-slow = 100',
    ], Section().logging.set_filters(slower=100))

    assert_lines([
        'log-filter = some',
    ], Section().logging.set_filters(include='some'))

    assert_lines([
        'log-drain = other',
    ], Section().logging.set_filters(exclude='other'))


def test_logging_add_logger(assert_lines):

    logging = Section().logging
    assert_lines([
        'worker-logger = my file:/home/here.log',
    ], logging.add_logger(logging.loggers.file('my', '/home/here.log'), for_single_worker=True))

    logging = Section().logging
    assert_lines([
        'logger = my socket:/home/here.sock',
    ], logging.add_logger(logging.loggers.socket('my', '/home/here.sock')))

    logging = Section().logging
    assert_lines([
        'logger = my syslog:myapp',
    ], logging.add_logger(logging.loggers.syslog('my', 'myapp')))

    logging = Section().logging
    assert_lines([
        'logger = my rsyslog:127.0.0.1:1111,myapp',
    ], logging.add_logger(logging.loggers.syslog('my', 'myapp', host='127.0.0.1:1111')))

    logging = Section().logging
    assert_lines([
        'logger = my redislog',
    ], logging.add_logger(logging.loggers.redis('my')))

    logging = Section().logging
    assert_lines([
        'logger = my mongodblog',
    ], logging.add_logger(logging.loggers.mongo('my')))

    logging = Section().logging
    assert_lines([
        'logger = my zeromq:tcp://192.168.173.18:9191',
    ], logging.add_logger(logging.loggers.zeromq('my', 'tcp://192.168.173.18:9191')))


def test_logging_add_logger_encoder(assert_lines):

    logging = Section().logging

    assert_lines([
        'worker-log-encoder = prefix -->',
        'worker-log-encoder = suffix <--',

    ], logging.add_logger_encoder([
        logging.encoders.prefix('-->'),
        logging.encoders.suffix('<--'),
    ], for_single_worker=True))

    logging = Section().logging
    enc_format = logging.encoders.format

    assert_lines([
        'log-encoder = format > ${msg} <:myfile',

    ], logging.add_logger_encoder([
        enc_format('> %s <' % enc_format.vars.MESSAGE),
    ], logger=logging.loggers.file('myfile', '/home/here.log')))



import pytest
import os

from uwsgiconf import Section, Configuration
from uwsgiconf.exceptions import ConfigurationError


def test_section_basics(ini, assert_lines):

    my_section = Section()

    assert_lines('automatically generated', my_section, stamp=True)

    my_section.set_basic_params(strict_config=True)

    assert_lines([
        'uwsgi',
        'strict',
    ], my_section)

    # unset bool
    my_section.set_basic_params(strict_config=False)
    assert_lines([
        'strict',
    ], my_section, assert_in=False)

    # successive unset bool
    my_section.set_basic_params(strict_config=False)

    assert_lines('plugins-list = true', Section().print_plugins())

    # __call__ -> set_basic_params
    assert_lines('workers = 10', Section().workers(count=10))

    # bogus basic params handling. test for no error
    Section(
        params_networking=None,
        params_nonexistent={'a': 'b'},
        dummy_key=1,
    )


def test_section_print(assert_lines):

    assert_lines('%[[37;49mAAA a%[[0m', Section(style_prints=True).print_out('a', indent='AAA '))
    assert_lines('= >   ===== variables', Section().print_variables())


def test_section_fallback(assert_lines):

    assert_lines('fallback-config = a/b.ini', Section().set_fallback('a/b.ini'))
    assert_lines('fallback-config = :here', Section().set_fallback(Section(name='here')))


def test_section_env(assert_lines):
    os.environ['Q'] = 'qq'

    assert_lines([
        'env = A=aa',
        'unenv = B',
        'env = Q=qq',
    ], Section().env('A', 'aa').env('Q').env('B', unset=True))


def test_section_include(assert_lines):

    assert_lines('ini = a/b.ini', Section().include('a/b.ini'))
    assert_lines('ini = :here', Section().include(Section(name='here')))


def test_section_derive_from(assert_lines):
    sec_base = Section(name='mine').workers.set_basic_params(count=3)

    assert_lines([
        '[mine]',
        'workers = 3',
    ], sec_base)

    sec1 = (
        Section.derive_from(sec_base).
            workers.set_basic_params(count=4).master_process.set_basic_params(enable=True))

    sec2 = (
        Section.derive_from(sec_base, name='other').
            workers.set_thread_params(enable=True))

    assert_lines([
        '[mine]',
        'workers = 4',
    ], sec1)

    assert_lines([
        '[other]',
        'workers = 3',
        'enable-threads = true',
    ], sec2)


def test_section_plugins(assert_lines):

    assert_lines([
        'plugin = plug',
        'autoload',
        'plugins-dir = /here',
        'plugins-dir = /there',
    ], Section().set_plugins_params(
        plugins='plug', search_dirs=['/here', '/there'], autoload=True
    ))


def test_plugin_init(assert_lines):
    assert_lines([
        'plugin = python34',

    ], Section(params_python={'version': 34, 'python_home': '/here'}))


def test_configuration(capsys, assert_lines):

    # basic params init
    assert_lines([
        'workers = 33',

    ], Section(params_workers=dict(count=33)))

    s1 = Section()
    s2 = 'some'

    with pytest.raises(ConfigurationError) as einfo:
        Configuration([s1, s2]).format()
    assert 'Section' in str(einfo.value)  # is a section

    s2 = Section()

    with pytest.raises(ConfigurationError) as einfo:
        Configuration([s1, s2]).format()
    assert 'unique' in str(einfo.value)  # has unique name

    s2.name = 'another'

    assert 'ini = :another' in Configuration([s1, s2], autoinclude_sections=True).format()

    assert Configuration([s1, s2]).print_ini()



from .alarms import Alarms
from .applications import Applications
from .caching import Caching
from .empire import Empire
from .locks import Locks
from .logging import Logging
from .main_process import MainProcess
from .master_process import MasterProcess
from .monitoring import Monitoring
from .networking import Networking
from .python import Python
from .queue import Queue
from .routing import Routing
from .spooler import Spooler
from .statics import Statics
from .workers import Workers, Cheapening



from uwsgiconf import Section


def test_locks_basics(assert_lines):

    assert_lines([
        'locks = 2',
    ], Section().locks.set_basic_params(count=2))

    assert_lines([
        'ftok = a',
    ], Section().locks.set_ipcsem_params(ftok='a'))

    assert_lines([
        'flock2 = /here',
    ], Section().locks.lock_file('/here', after_setup=True))




from operator import attrgetter

from ..base import OptionsGroup, ParametrizedValue
from ..utils import listify, filter_locals, make_key_val_string


class AlarmType(ParametrizedValue):

    def __init__(self, alias, *args, **kwargs):
        self.alias = alias or ''
        super(AlarmType, self).__init__(*args)


class AlarmCommand(AlarmType):
    """Run a shell command, passing info into its stdin."""

    name = 'cmd'

    def __init__(self, alias, command):
        super(AlarmCommand, self).__init__(alias, command)


class AlarmSignal(AlarmType):
    """Raise an uWSGI signal."""

    name = 'signal'

    def __init__(self, alias, sig_num):
        super(AlarmSignal, self).__init__(alias, sig_num)


class AlarmLog(AlarmType):
    """Print line into log."""

    name = 'log'

    def __init__(self, alias):
        super(AlarmLog, self).__init__(alias)


class AlarmMule(AlarmType):
    """Send info to a mule waiting for messages."""

    name = 'mule'

    def __init__(self, alias, mule_id):
        super(AlarmMule, self).__init__(alias, mule_id)


class AlarmCurl(AlarmType):
    """Send info to a cURL-able URL."""

    name = 'curl'
    plugin = 'alarm_curl'
    args_joiner = ';'

    def __init__(
            self, alias, url, method=None, ssl=None, ssl_insecure=None,
            auth_user=None, auth_pass=None,
            timeout=None, conn_timeout=None,
            mail_from=None, mail_to=None, subject=None):

        opts = make_key_val_string(
            filter_locals(locals(), drop=['alias', 'url']),
            bool_keys=['ssl', 'ssl_insecure'],
            items_separator=self.args_joiner,
        )
        super(AlarmCurl, self).__init__(alias, url, opts)


class AlarmXmpp(AlarmType):
    """Send info via XMPP/jabber."""

    name = 'xmpp'
    plugin = 'alarm_xmpp'
    args_joiner = ';'

    def __init__(self, alias, jid, password, recipients):
        super(AlarmXmpp, self).__init__(alias, jid, password, ','.join(listify(recipients)))


class Alarms(OptionsGroup):
    """Alarms.

    This subsystem allows the developer/sysadmin to "announce"
    special conditions of an app via various channels.

    * http://uwsgi-docs.readthedocs.io/en/latest/AlarmSubsystem.html

    """

    class alarm_types(object):
        """Alarm types available for ``.register_alarm()``."""

        command = AlarmCommand
        curl = AlarmCurl
        log = AlarmLog
        mule = AlarmMule
        signal = AlarmSignal
        xmpp = AlarmXmpp

    def __init__(self, *args, **kwargs):
        self._alarms = []
        super(Alarms, self).__init__(*args, **kwargs)

    def set_basic_params(self, msg_size=None, cheap=None, anti_loop_timeout=None):
        """
        :param int msg_size: Set the max size of an alarm message in bytes. Default: 8192.

        :param bool cheap: Use main alarm thread rather than create dedicated
            threads for curl-based alarms

        :param int anti_loop_timeout: Tune the anti-loop alarm system. Default: 3 seconds.

        """
        self._set('alarm-msg-size', msg_size)
        self._set('alarm-cheap', cheap, cast=bool)
        self._set('alarm-freq', anti_loop_timeout)

        return self._section

    def print_alarms(self):
        """Print out enabled alarms."""

        self._set('alarm-list', True, cast=bool)

        return self._section

    def register_alarm(self, alarm):
        """Register (create) an alarm.

        :param AlarmType|list[AlarmType] alarm: Alarm.

        """
        for alarm in listify(alarm):
            if alarm not in self._alarms:
                self._set('alarm', alarm, multi=True)
                self._alarms.append(alarm)

        return self._section

    def alarm_on_log(self, alarm, matcher, skip=False):
        """Raise (or skip) the specified alarm when a log line matches the specified regexp.

        :param AlarmType|list[AlarmType] alarm: Alarm.

        :param str|unicode matcher: Regular expression to match log line.

        :param bool skip:

        """
        self.register_alarm(alarm)

        value = '%s %s' % (
            ','.join(map(attrgetter('alias'), listify(alarm))),
            matcher)

        self._set('not-alarm-log' if skip else 'alarm-log', value)

        return self._section

    def alarm_on_fd_ready(self, alarm, fd, message, byte_count=None):
        """Triggers the alarm when the specified file descriptor is ready for read.

        This is really useful for integration with the Linux eventfd() facility.
        Pretty low-level and the basis of most of the alarm plugins.

        * http://uwsgi-docs.readthedocs.io/en/latest/Changelog-1.9.7.html#alarm-fd

        :param AlarmType|list[AlarmType] alarm: Alarm.

        :param str|unicode fd: File descriptor.

        :param str|unicode message: Message to send.

        :param int byte_count: Files to read. Default: 1 byte.

            .. note:: For ``eventfd`` set 8.

        """
        self.register_alarm(alarm)

        value = fd

        if byte_count:
            value += ':%s' % byte_count

        value += ' %s' % message

        for alarm in listify(alarm):
            self._set('alarm-fd', '%s %s' % (alarm.alias, value), multi=True)

        return self._section

    def alarm_on_backlog_full(self, alarm):
        """Raise the specified alarm when the socket backlog queue is full.

        :param AlarmType|list[AlarmType] alarm: Alarm.
        """
        self.register_alarm(alarm)

        for alarm in listify(alarm):
            self._set('alarm-backlog', alarm.alias, multi=True)

        return self._section

    def alarm_on_segfault(self, alarm):
        """Raise the specified alarm when the segmentation fault handler is executed.

        Sends a backtrace.

        :param AlarmType|list[AlarmType] alarm: Alarm.
        """
        self.register_alarm(alarm)

        for alarm in listify(alarm):
            self._set('alarm-segfault', alarm.alias, multi=True)

        return self._section



import pytest

from uwsgiconf import Section
from uwsgiconf.exceptions import ConfigurationError


def test_plugin_python_basics(assert_lines):

    def get_plugin_group():
        section = Section()
        return section.python

    assert_lines([
        'enable-threads = true',
    ], get_plugin_group().set_basic_params(enable_threads=True))

    assert_lines([
        'pyargv = a b',
    ], get_plugin_group().set_app_args('a', 'b'))

    assert_lines([
        'wsgi = myapp.py',
    ], get_plugin_group().set_wsgi_params(module='myapp.py'))

    assert_lines([
        'wsgi-file = some/file.py',
    ], get_plugin_group().set_wsgi_params(module='some/file.py'))

    assert_lines([
        'eval = print("!")',
    ], get_plugin_group().eval_wsgi_entrypoint(code='print("!")'))

    assert_lines([
        'py-auto-reload = 10',
    ], get_plugin_group().set_autoreload_params(scan_interval=10))

    assert_lines([
        'post-pymodule-alias = ali=here/file.py',
    ], get_plugin_group().register_module_alias('ali', 'here/file.py', after_init=True))

    assert_lines([
        'post-pymodule-alias = ali=here/file.py',
    ], get_plugin_group().register_module_alias('ali', 'here/file.py', after_init=True))

    with pytest.raises(ConfigurationError) as einfo:
        assert_lines('', get_plugin_group().import_module('some.module', shared=True, into_spooler=True))
    assert 'set both' in str(einfo.value)

    assert_lines([
        'python-import = some/module.py',
    ], get_plugin_group().import_module('some/module.py'))

    assert_lines([
        'shared-python-import = some/sharemodule.py',
    ], get_plugin_group().import_module('some/sharemodule.py', shared=True))

    assert_lines([
        'spooler-python-import = some/forspooler.py',
    ], get_plugin_group().import_module('some/forspooler.py', into_spooler=True))

    assert_lines([
        'pyrun = runit.py',
    ], get_plugin_group().run_module('runit.py'))



from uwsgiconf import Section


def test_main_process_basics(assert_lines):

    assert_lines([
        'touch-reload = /here/a',
    ], Section().main_process.set_basic_params(touch_reload='/here/a'))

    assert_lines([
        'daemonize = /here/app.log',
    ], Section().main_process.daemonize('/here/app.log'))

    assert_lines([
        'daemonize2 = /here/app.log',
    ], Section().main_process.daemonize('/here/app.log', after_app_loading=True))

    assert_lines([
        'pidfile = /here/file',
    ], Section().main_process.set_pid_file('/here/file'))

    assert_lines([
        'safe-pidfile2 = /here/file',
    ], Section().main_process.set_pid_file(
        '/here/file', before_priv_drop=False, safe=True))

    assert_lines([
        'auto-procname = true',
    ], Section().main_process.set_naming_params(autonaming=True))

    assert_lines([
        'uid = 1001',
    ], Section().main_process.set_owner_params(uid=1001))

    assert_lines([
        'exec-asap = date',
    ], Section().main_process.run_command_on_event('date'))

    assert_lines([
        'touch-exec = myfile.txt date',
    ], Section().main_process.run_command_on_touch('date', 'myfile.txt'))

    assert_lines([
        'skip-atexit = true',
    ], Section().main_process.set_on_exit_params(skip_hooks=True))


def test_main_process_hooks(assert_lines):

    section = Section()

    prc = section.main_process
    asap = prc.phases.ASAP

    prc.set_hook(asap, prc.actions.mount('/proc', 'proc', 'none'))
    prc.set_hook(asap, prc.actions.mount('/proc', flags=['rec', 'detach']))
    prc.set_hook(asap, prc.actions.execute('cat /proc/self/mounts'))
    prc.set_hook(asap, prc.actions.call('uwsgi_log application has been loaded'))
    prc.set_hook(asap, prc.actions.call('putenv PATH=bin:$(PATH)', arg_int=True))
    prc.set_hook(asap, prc.actions.call('some', honour_exit_status=True))
    prc.set_hook(asap, prc.actions.dir_change('/here'))
    prc.set_hook(asap, prc.actions.exit())
    prc.set_hook(prc.phases.APP_LOAD_PRE, prc.actions.exit(10))
    prc.set_hook(asap, prc.actions.printout('bingo-bongo'))
    prc.set_hook(asap, prc.actions.file_write('/here/a.txt', 'sometext', append=True, newline=True))
    prc.set_hook(asap, prc.actions.fifo_write('/here/b', '10', wait=True))
    prc.set_hook(asap, prc.actions.unlink('/here/d'))
    prc.set_hook(asap, prc.actions.alarm('myal', 'bang'))
    prc.set_hook(asap, prc.actions.set_host_name('newname'))
    prc.set_hook(asap, prc.actions.file_create('/here/a.txt'))
    prc.set_hook(asap, prc.actions.dir_create('/here/there'))

    assert_lines([
        'hook-asap = mount:proc none /proc',
        'hook-asap = umount:/proc rec,detach',
        'hook-asap = print:bingo-bongo',
        'hook-asap = exec:cat /proc/self/mounts',
        'hook-asap = call:uwsgi_log application has been loaded',
        'hook-asap = callint:putenv PATH=bin:$(PATH)',
        'hook-asap = callret:some',
        'hook-asap = cd:/here',
        'hook-asap = exit:',
        'hook-pre-app = exit:10',
        'hook-asap = appendn:/here/a.txt sometext',
        'hook-asap = spinningfifo:/here/b 10',
        'hook-asap = unlink:/here/d',
        'hook-asap = alarm:myal bang',
        'hook-asap = hostname:newname',
        'hook-asap = create:/here/a.txt',
        'hook-asap = mkdir:/here/there',
    ], section)

    assert_lines([
        'hook-touch = /that do',
    ], Section().main_process.set_hook_touch('/that', 'do'))

    assert_lines([
        'after-request-hook = cfunc',
    ], Section().main_process.set_hook_after_request('cfunc'))



from ..base import OptionsGroup


class Empire(OptionsGroup):
    """Emperor and his vassals.

    If you need to deploy a big number of apps on a single server,
    or a group of servers, the Emperor mode is just the ticket.

    * http://uwsgi-docs.readthedocs.io/en/latest/Emperor.html

    """

    def set_emperor_params(
            self, vassals_home=None, name=None, scan_interval=None, pid_file=None, spawn_asap=None,
            stats_address=None, trigger_socket=None, links_no_follow=None):
        """

        :param str|unicode|list[str|unicode] vassals_home: Set vassals home and enable Emperor mode.

        :param str|unicode name: Set the Emperor process name.

        :param int scan_interval: Set the Emperor scan frequency. Default: 3 seconds.

        :param str|unicode pid_file: Write the Emperor pid in the specified file.

        :param bool spawn_asap: Spawn the Emperor as soon as possible.

        :param str|unicode stats_address: Run the Emperor stats server on specified address.

        :param str|unicode trigger_socket: Enable the Emperor trigger socket.

        :param bool links_no_follow: Do not follow symlinks when checking for mtime.

        """
        self._set('emperor', vassals_home, multi=True)
        self._set('emperor-procname', name)
        self._set('emperor-freq', scan_interval)
        self._set('emperor-pidfile', pid_file)
        self._set('early-emperor', spawn_asap, cast=bool)
        self._set('emperor-stats-server', stats_address)
        self._set('emperor-trigger-socket', trigger_socket)
        self._set('emperor-nofollow', links_no_follow)

        return self._section

    def print_monitors(self):
        """Print out enabled imperial monitors."""

        self._set('imperial-monitor-list', True, cast=bool)

        return self._section

    def set_emperor_command_params(
            self, command_socket=None,
            wait_for_command=None, wait_for_command_exclude=None):
        """Emperor commands related parameters.

        * http://uwsgi-docs.readthedocs.io/en/latest/tutorials/EmperorSubscriptions.html

        :param str|unicode command_socket: Enable the Emperor command socket.

        :param bool wait_for_command: Always wait for a 'spawn' Emperor command before starting a vassal.

        :param str|unicode|list[str|unicode] wait_for_command_exclude: Vassals that will ignore ``wait_for_command``.

        """
        self._set('emperor-command-socket', command_socket)
        self._set('emperor-wait-for-command', wait_for_command, cast=bool)
        self._set('emperor-wait-for-command-ignore', wait_for_command_exclude, multi=True)

        return self._section

    def set_vassals_wrapper_params(self, wrapper=None, overrides=None, fallbacks=None):
        """Binary wrapper for vassals parameters.

        :param str|unicode wrapper: Set a binary wrapper for vassals.

        :param str|unicode|list[str|unicode] overrides: Set a binary wrapper for vassals to try before the default one

        :param str|unicode|list[str|unicode] fallbacks: Set a binary wrapper for vassals to try as a last resort.
            Allows you to specify an alternative binary to execute when running a vassal
            and the default binary_path is not found (or returns an error).

        """
        self._set('emperor-wrapper', wrapper)
        self._set('emperor-wrapper-override', overrides, multi=True)
        self._set('emperor-wrapper-fallback', fallbacks, multi=True)

        return self._section

    def set_throttle_params(self, level=None, level_max=None):
        """Throttling options.

        * http://uwsgi-docs.readthedocs.io/en/latest/Emperor.html#throttling
        * http://uwsgi-docs.readthedocs.io/en/latest/Emperor.html#loyalty

        :param int level: Set throttling level (in milliseconds) for bad behaving vassals. Default: 1000.

        :param int level_max: Set maximum throttling level (in milliseconds)
            for bad behaving vassals. Default: 3 minutes.

        """
        self._set('emperor-throttle', level)
        self._set('emperor-max-throttle', level_max)

        return self._section

    def set_tolerance_params(self, for_heartbeat=None, for_cursed_vassals=None):
        """Various tolerance options.

        :param int for_heartbeat: Set the Emperor tolerance about heartbeats.

            * http://uwsgi-docs.readthedocs.io/en/latest/Emperor.html#heartbeat-system

        :param int for_cursed_vassals: Set the Emperor tolerance about cursed vassals.

            * http://uwsgi-docs.readthedocs.io/en/latest/Emperor.html#blacklist-system

        """
        self._set('emperor-required-heartbeat', for_heartbeat)
        self._set('emperor-curse-tolerance', for_cursed_vassals)

        return self._section

    def set_mode_tyrant_params(self, enable=None, links_no_follow=None, use_initgroups=None):
        """Tyrant mode (secure multi-user hosting).

        In Tyrant mode the Emperor will run the vassal using the UID/GID of the vassal
        configuration file.

        * http://uwsgi-docs.readthedocs.io/en/latest/Emperor.html#tyrant-mode-secure-multi-user-hosting

        :param enable: Puts the Emperor in Tyrant mode.

        :param bool links_no_follow: Do not follow symlinks when checking for uid/gid in Tyrant mode.

        :param bool use_initgroups: Add additional groups set via initgroups() in Tyrant mode.

        """
        self._set('emperor-tyrant', enable, cast=bool)
        self._set('emperor-tyrant-nofollow', links_no_follow, cast=bool)
        self._set('emperor-tyrant-initgroups', use_initgroups, cast=bool)

        return self._section

    def set_mode_broodlord_params(
            self, zerg_count=None,
            vassal_overload_sos_interval=None, vassal_backlog_items_sos=None):
        """This mode is a way for a vassal to ask for reinforcements to the Emperor.

        Reinforcements are new vassals spawned on demand generally bound on the same socket.

        .. warning:: If you are looking for a way to dynamically adapt the number
            of workers of an instance, check the Cheaper subsystem - adaptive process spawning mode.

            *Broodlord mode is for spawning totally new instances.*

        :param int zerg_count: Maximum number of zergs to spawn.

        :param int vassal_overload_sos_interval: Ask emperor for reinforcement when overloaded.
            Accepts the number of seconds to wait between asking for a new reinforcements.

        :param int vassal_backlog_items_sos: Ask emperor for sos if backlog queue has more
            items than the value specified

        """
        self._set('emperor-broodlord', zerg_count)
        self._set('vassal-sos', vassal_overload_sos_interval)
        self._set('vassal-sos-backlog', vassal_backlog_items_sos)

        return self._section



from uwsgiconf import Section


def test_spooler_basics(assert_lines):

    assert_lines([
        'touch-spoolers-reload = /here/a',
    ], Section().spooler.set_basic_params(touch_reload='/here/a'))

    assert_lines([
        'spooler = home/one',
        'spooler = home/two',
    ], Section().spooler.add(['home/one', 'home/two']))

    assert_lines([
        'spooler-external = home/two',
    ], Section().spooler.add('home/two', external=True))



from ..base import OptionsGroup


class Spooler(OptionsGroup):
    """Spooler.

    .. note:: Supported on: Perl, Python, Ruby.

    .. note:: Be sure the ``spooler`` plugin is loaded in your instance,
        but generally it is built in by default.

    The Spooler is a queue manager built into uWSGI that works like a printing/mail system.
    You can enqueue massive sending of emails, image processing, video encoding, etc.
    and let the spooler do the hard work in background while your users
    get their requests served by normal workers.

    http://uwsgi-docs.readthedocs.io/en/latest/Spooler.html

    """

    def set_basic_params(
            self, touch_reload=None, quiet=None, process_count=None, max_tasks=None,
            order_tasks=None, harakiri=None, change_dir=None, poll_interval=None, signal_as_task=None,
            cheap=None):
        """

        :param str|unicode|list touch_reload: reload spoolers if the specified file is modified/touched

        :param bool quiet: Do not be verbose with spooler tasks.

        :param int process_count: Set the number of processes for spoolers.

        :param int max_tasks: Set the maximum number of tasks to run before recycling
            a spooler (to help alleviate memory leaks).

        :param int order_tasks: Try to order the execution of spooler tasks (uses scandir instead of readdir).

        :param int harakiri: Set harakiri timeout for spooler tasks.

        :param str|unicode change_dir: chdir() to specified directory before each spooler task.

        :param int poll_interval: Spooler poll frequency in seconds. Default: 30.

        :param bool signal_as_task: Treat signal events as tasks in spooler,
            combine used with ``spooler-max-tasks``. Enable this, spooler will treat signal
            events as task. run signal handler will also increase the spooler task count.

        :param bool cheap: Use spooler cheap mode.

        """
        self._set('touch-spoolers-reload', touch_reload, multi=True)
        self._set('spooler-quiets', quiet, cast=bool)
        self._set('spooler-processes', process_count)
        self._set('spooler-max-tasks', max_tasks)
        self._set('spooler-ordered', order_tasks, cast=bool)
        self._set('spooler-harakiri', harakiri)
        self._set('spooler-chdir', change_dir)
        self._set('spooler-frequency', poll_interval)
        self._set('spooler-signal-as-task', signal_as_task, cast=bool)
        self._set('spooler-cheap', cheap, cast=bool)

        return self._section

    def add(self, work_dir, external=False):
        """run a spooler on the specified directory

        :param work_dir:

        :param bool external: map spoolers requests to a spooler directory managed by an external instance

        """
        command = 'spooler'

        if external:
            command += '-external'

        self._set(command, work_dir, multi=True)

        return self._section



from uwsgiconf import Section


def test_queue_basics(assert_lines):

    assert_lines([
        'queue = 100',
        'queue-blocksize = 131072',
    ], Section().queue.enable(100, block_size=131072))



import pytest

from uwsgiconf import Section
from uwsgiconf.exceptions import ConfigurationError


def test_master_process_basics(assert_lines):

    assert_lines([
        'master = true',
    ], Section().master_process.set_basic_params(enable=True))

    assert_lines([
        'reload-mercy = 10',
    ], Section().master_process.set_reload_params(mercy=10))

    assert_lines([
        'idle = 10',
    ], Section().master_process.set_idle_params(timeout=10))


def test_master_attach_process_classic(assert_lines):

    assert_lines([
        'smart-attach-daemon2 = /here/a.pid',
    ], Section().master_process.attach_process_classic('/here/a.pid', background=False))

    assert_lines([
        'smart-attach-daemon = /here/a.pid',
    ], Section().master_process.attach_process_classic('/here/a.pid', background=True))

    with pytest.raises(ConfigurationError):
        assert_lines([
            '',
        ], Section().master_process.attach_process_classic('command', background=True))

    assert_lines([
        'attach-daemon = command',
    ], Section().master_process.attach_process_classic('command', background=False))

    assert_lines([
        'attach-control-daemon = command',
    ], Section().master_process.attach_process_classic(
        'command', background=False, control=True))


def test_master_attach_process(assert_lines):

    assert_lines([
        'attach-daemon2 = cmd=date',
    ], Section().master_process.attach_process('date'))

    assert_lines([
        'legion-attach-daemon2 = cmd=date',
    ], Section().master_process.attach_process('date', for_legion=True))

    assert_lines([
        'attach-daemon2 = cmd=date,pidfile=/here/my.pid,control=1,touch=/here/one',
    ], Section().master_process.attach_process('date', control=True, pidfile='/here/my.pid', touch_reload='/here/one'))

    assert_lines([
        'attach-daemon2 = cmd=date,touch=/here/one;/there/two',
    ], Section().master_process.attach_process('date', touch_reload=['/here/one', '/there/two']))


def test_cron(assert_lines):

    assert_lines([
        'cron2 = torrt walk',
    ], Section().master_process.add_cron_task('torrt walk'))

    assert_lines([
        'cron2 = week=1-3,hour=2,minute=-10,harakiri=10,legion=first,unique=1 some',
    ], Section().master_process.add_cron_task(
        'some', hour=2, minute=-10, weekday='1-3', harakiri=10, unique=True, legion='first'
    ))



from uwsgiconf import Section


def test_caching_basics(assert_lines):

    assert_lines([
        'cache-no-expire = true',
    ], Section().caching.set_basic_params(no_expire=True))

    assert_lines([
        'add-cache-item = a=b',
    ], Section().caching.add_item('a', 'b'))

    assert_lines([
        'add-cache-item = some a=b',
    ], Section().caching.add_item('a', 'b', cache_name='some'))

    assert_lines([
        'load-file-in-cache = /home/some.a',
    ], Section().caching.add_file('/home/some.a'))

    assert_lines([
        'load-file-in-cache-gzip = /home/some.a',
    ], Section().caching.add_file('/home/some.a', gzip=True))

    assert_lines([
        'load-file-in-cache = some /home/some.a',
    ], Section().caching.add_file('/home/some.a', cache_name='some'))

    assert_lines([
        'cache2 = name=mine,maxitems=333',
    ], Section().caching.add_cache('mine', 333))






def format_print_text(text, color_fg=None, color_bg=None):
    """Format given text using ANSI formatting escape sequences.

    Could be useful gfor print command.

    :param str|unicode text:
    :param str|unicode color_fg: text (foreground) color
    :param str|unicode color_bg: text (background) color
    :rtype: str|unicode
    """
    from .config import Section

    color_fg = {

        'black': '30',
        'red': '31',
        'reder': '91',
        'green': '32',
        'greener': '92',
        'yellow': '33',
        'yellower': '93',
        'blue': '34',
        'bluer': '94',
        'magenta': '35',
        'magenter': '95',
        'cyan': '36',
        'cyaner': '96',
        'gray': '37',
        'grayer': '90',
        'white': '97',

    }.get(color_fg, '39')

    color_bg = {

        'black': '40',
        'red': '41',
        'reder': '101',
        'green': '42',
        'greener': '102',
        'yellow': '43',
        'yellower': '103',
        'blue': '44',
        'bluer': '104',
        'magenta': '45',
        'magenter': '105',
        'cyan': '46',
        'cyaner': '106',
        'gray': '47',
        'grayer': '100',
        'white': '107',

    }.get(color_bg, '49')

    mod = ';'.join([color_fg, color_bg])

    text = '%(esc)s[%(mod)sm%(value)s%(end)s' % {
        'esc': Section.vars.FORMAT_ESCAPE,
        'end': Section.vars.FORMAT_END,
        'mod': mod,
        'value': text,
    }

    return text


class FormatterBase(object):
    """Base class for configuration formatters."""

    def __init__(self, sections):
        self.sections = sections

    def iter_options(self):
        """Iterates configuration sections groups options."""
        for section in self.sections:
            name = str(section)
            for key, value in section._get_options():
                yield name, key, value

    def format(self):
        raise NotImplementedError()  # pragma: nocover


class IniFormatter(FormatterBase):
    """Translates a configuration as INI file."""

    def format(self):
        lines = []
        last_section = ''

        for section_name, key, value in self.iter_options():

            if section_name != last_section:
                lines.append('\n[%s]' % section_name)
                last_section = section_name

            lines.append('%s = %s' % (key, str(value).strip()))

        lines = '\n'.join(lines)
        return lines



from functools import partial

from ..base import OptionsGroup
from ..utils import listify


class MuleFarm(object):
    """Represents a mule farm."""

    def __init__(self, name, mule_numbers):
        """
        :param str|unicode name: Farm alias.

        :param int|list[int] mule_numbers: Total mules on farm count,
            or a list of mule numbers.

        """
        # todo http://uwsgi-docs.readthedocs.io/en/latest/Signals.html#signals-targets
        self.name = name
        self.mule_numbers = mule_numbers

    def __str__(self):
        return '%s:%s' % (self.name, ','.join(map(str, self.mule_numbers)))


class Workers(OptionsGroup):
    """Workers aka [working] processes."""

    mule_farm = MuleFarm

    def set_basic_params(
            self, count=None, touch_reload=None, touch_chain_reload=None, zombie_reaper=None,
            limit_addr_space=None, limit_count=None, cpu_affinity=None):
        """

        :param int count: Spawn the specified number of workers (processes).
            Set the number of workers for preforking mode.
            This is the base for easy and safe concurrency in your app.
            More workers you add, more concurrent requests you can manage.

            Each worker corresponds to a system process, so it consumes memory,
            choose carefully the right number. You can easily drop your system
            to its knees by setting a too high value.

            Setting ``workers`` to a ridiculously high number will **not**
            magically make your application web scale - quite the contrary.

        :param str|list touch_reload: Trigger reload of (and only) workers
            if the specified file is modified/touched.

        :param str|list touch_chain_reload: Trigger chain workers reload on file touch.
            When in lazy/lazy_apps mode, you can simply destroy a worker to force it to reload the application code.
            A new reloading system named "chain reload", allows you to reload one worker at time
            (opposed to the standard way where all of the workers are destroyed in bulk)

            * http://uwsgi-docs.readthedocs.io/en/latest/articles/TheArtOfGracefulReloading.html#chain-reloading-lazy-apps

        :param bool zombie_reaper: Call waitpid(-1,...) after each request to get rid of zombies.
            Enables reaper mode. After each request the server will call ``waitpid(-1)``
            to get rid of zombie processes.
            If you spawn subprocesses in your app and you happen to end up with zombie processes
            all over the place you can enable this option. (It really would be better
            if you could fix your application's process spawning usage though.)

        :param int limit_addr_space: Limit process address space (vsz) (in megabytes)
            Limits the address space usage of each uWSGI (worker) process using POSIX/UNIX ``setrlimit()``.
            For example, ``limit-as 256`` will disallow uWSGI processes to grow over 256MB of address space.
            Address space is the virtual memory a process has access to. It does *not* correspond to physical memory.
            Read and understand this page before enabling this option: http://en.wikipedia.org/wiki/Virtual_memory

        :param int limit_count: Limit the number of spawnable processes.

        :param int cpu_affinity: number of cores for each worker (Linux only)
            Set the number of cores (CPUs) to allocate to each worker process.

            * 4 workers, 4 CPUs, affinity is 1,
                each worker is allocated one CPU.

            * 4 workers, 2 CPUs, affinity is 1,
                workers get one CPU each (0; 1; 0; 1).

            * 4 workers, 4 CPUs, affinity is 2,
                workers get two CPUs each in a round-robin fashion (0, 1; 2, 3; 0, 1; 2; 3).

            * 8 workers, 4 CPUs, affinity is 3,
                workers get three CPUs each in a round-robin fashion
                (0, 1, 2; 3, 0, 1; 2, 3, 0; 1, 2, 3; 0, 1, 2; 3, 0, 1; 2, 3, 0; 1, 2, 3).

        """
        # todo map socket to a worker - map-socket = 0:1
        # networking.register_socket - workers_binding

        self.set_count_auto(count)

        self._set('touch-workers-reload', touch_reload, multi=True)
        self._set('touch-chain-reload', touch_chain_reload, multi=True)

        self._set('reaper', zombie_reaper, cast=bool)

        self._set('limit-as', limit_addr_space)
        self._set('limit-nproc', limit_count)

        self._set('cpu-affinity', cpu_affinity)

        return self._section

    def run_command_as_worker(self, command, after_post_fork_hook=False):
        """Run the specified command as worker.

        :param str|unicode command:

        :param bool after_post_fork_hook: Whether to run it after `post_fork` hook.

        """
        self._set('worker-exec2' if after_post_fork_hook else 'worker-exec', command, multi=True)

        return self._section

    def set_count_auto(self, count=None):
        """Sets workers count.

        By default sets it to detected number of available cores

        :param int count:
        """
        count = count or self._section.vars.CPU_CORES

        self._set('workers', count)

        return self._section

    def set_thread_params(
            self, enable=None, count=None, count_offload=None, stack_size=None, no_wait=None):
        """Sets threads related params.

        :param bool enable: Enable threads in the embedded languages.
            This will allow to spawn threads in your app.
            Threads will simply *not work* if this option is not enabled. There will likely be no error,
            just no execution of your thread code.

        :param int count: Run each worker in prethreaded mode with the specified number
            of threads per worker.

            .. warning:: Do not use with ``gevent``.

            .. note:: Enables threads automatically.

        :param int count_offload: Set the number of threads (per-worker) to spawn
            for offloading. Default: 0.

            These threads run such tasks in a non-blocking/evented way allowing
            for a huge amount of concurrency. Various components of the uWSGI stack
            are offload-friendly.

            .. note:: Try to set it to the number of CPU cores to take advantage of SMP.

            * http://uwsgi-docs.readthedocs.io/en/latest/OffloadSubsystem.html

        :param int stack_size: Set threads stacksize.

        :param bool no_wait: Do not wait for threads cancellation on quit/reload.

        """
        self._set('enable-threads', enable, cast=bool)
        self._set('no-threads-wait', no_wait, cast=bool)
        self._set('threads', count)
        self._set('offload-threads', count_offload)

        if count:
            self._section.print_out('Threads per worker: %s' % count)

        self._set('threads-stacksize', stack_size)

        return self._section

    def set_mules_params(
            self, mules=None, touch_reload=None, harakiri_timeout=None, farms=None, reload_mercy=None,
            msg_buffer=None, msg_buffer_recv=None):
        """Sets mules related params.

        http://uwsgi.readthedocs.io/en/latest/Mules.html

        Mules are worker processes living in the uWSGI stack but not reachable via socket connections,
        that can be used as a generic subsystem to offload tasks.

        :param int|list mules: Add the specified mules or number of mules.

        :param str|list touch_reload: Reload mules if the specified file is modified/touched.

        :param int harakiri_timeout: Set harakiri timeout for mule tasks.

        :param list[MuleFarm] farms: Mule farms list.

            Examples:
                * cls_mule_farm('first', 2)
                * cls_mule_farm('first', [4, 5])

        :param int reload_mercy: Set the maximum time (in seconds) a mule can take
            to reload/shutdown. Default: 60.

        :param int msg_buffer: Set mule message buffer size (bytes) given
            for mule message queue.

        :param int msg_buffer: Set mule message recv buffer size (bytes).

        """
        farms = farms or []

        next_mule_number = 1
        farm_mules_count = 0

        for farm in farms:

            if isinstance(farm.mule_numbers, int):
                farm.mule_numbers = list(range(next_mule_number, next_mule_number + farm.mule_numbers))
                next_mule_number = farm.mule_numbers[-1] + 1

            farm_mules_count += len(farm.mule_numbers)

            self._set('farm', farm, multi=True)

        if mules is None and farm_mules_count:
            mules = farm_mules_count

        if isinstance(mules, int):
            self._set('mules', mules)

        elif isinstance(mules, list):

            for mule in mules:
                self._set('mule', mule, multi=True)

        self._set('touch-mules-reload', touch_reload, multi=True)

        self._set('mule-harakiri', harakiri_timeout)
        self._set('mule-reload-mercy', reload_mercy)

        self._set('mule-msg-size', msg_buffer)
        self._set('mule-msg-recv-size', msg_buffer_recv)

        return self._section

    def set_reload_params(
            self, min_lifetime=None, max_lifetime=None,
            max_requests=None, max_requests_delta=None,
            max_addr_space=None, max_rss=None, max_uss=None, max_pss=None,
            max_addr_space_forced=None, max_rss_forced=None,
            mercy=None):
        """Sets workers reload parameters.

        :param int min_lifetime: A worker cannot be destroyed/reloaded unless it has been alive
            for N seconds (default 60). This is an anti-fork-bomb measure.
            Since 1.9

        :param int max_lifetime: Reload workers after this many seconds. Disabled by default.
            Since 1.9

        :param int max_requests: Reload workers after the specified amount of managed
            requests (avoid memory leaks).
            When a worker reaches this number of requests it will get recycled (killed and restarted).
            You can use this option to "dumb fight" memory leaks.

            Also take a look at the ``reload-on-as`` and ``reload-on-rss`` options
            as they are more useful for memory leaks.

            .. warning:: The default min-worker-lifetime 60 seconds takes priority over `max-requests`.

            Do not use with benchmarking as you'll get stalls
            such as `worker respawning too fast !!! i have to sleep a bit (2 seconds)...`

        :param int max_requests_delta: Add (worker_id * delta) to the max_requests value of each worker.

        :param int max_addr_space: Reload a worker if its address space usage is higher
            than the specified value in megabytes.

        :param int max_rss: Reload a worker if its physical unshared memory (resident set size) is higher
            than the specified value (in megabytes).

        :param int max_uss: Reload a worker if Unique Set Size is higher
            than the specified value in megabytes.

            .. note:: Linux only.

        :param int max_pss: Reload a worker if Proportional Set Size is higher
            than the specified value in megabytes.

            .. note:: Linux only.

        :param int max_addr_space_forced: Force the master to reload a worker if its address space is higher
            than specified megabytes (in megabytes).

        :param int max_rss_forced: Force the master to reload a worker
            if its resident set size memory is higher than specified in megabytes.

        :param int mercy: Set the maximum time (in seconds) a worker can take
            before reload/shutdown. Default: 60.

        """
        self._set('max-requests', max_requests)
        self._set('max-requests-delta', max_requests_delta)

        self._set('min-worker-lifetime', min_lifetime)
        self._set('max-worker-lifetime', max_lifetime)

        self._set('reload-on-as', max_addr_space)
        self._set('reload-on-rss', max_rss)
        self._set('reload-on-uss', max_uss)
        self._set('reload-on-pss', max_pss)

        self._set('evil-reload-on-as', max_addr_space_forced)
        self._set('evil-reload-on-rss', max_rss_forced)

        self._set('worker-reload-mercy', mercy)

        return self._section

    def set_reload_on_exception_params(self, do_reload=None, etype=None, evalue=None, erepr=None):
        """Sets workers reload on exceptions parameters.

        :param bool do_reload: Reload a worker when an exception is raised.

        :param str etype: Reload a worker when a specific exception type is raised.

        :param str evalue: Reload a worker when a specific exception value is raised.

        :param str erepr: Reload a worker when a specific exception type+value (language-specific) is raised.

        """
        self._set('reload-on-exception', do_reload, cast=bool)
        self._set('reload-on-exception-type', etype)
        self._set('reload-on-exception-value', evalue)
        self._set('reload-on-exception-repr', erepr)

        return self._section

    def set_harakiri_params(self, timeout=None, verbose=None, disable_for_arh=None):
        """Sets workers harakiri parameters.

        :param timeout: Harakiri timeout in seconds.
            Every request that will take longer than the seconds specified
            in the harakiri timeout will be dropped and the corresponding
            worker is thereafter recycled.

        :param verbose: Harakiri verbose mode.
            When a request is killed by Harakiri you will get a message in the uWSGI log.
            Enabling this option will print additional info (for example,
            the current syscall will be reported on Linux platforms).

        :param disable_for_arh: Disallow Harakiri killings during after-request hook methods.

        """
        self._set('harakiri', timeout)
        self._set('harakiri-verbose', verbose)
        self._set('harakiri-no-arh', disable_for_arh)

        return self._section

    def set_zerg_server_params(self, socket, clients_socket_pool=None):
        """Zerg mode. Zerg server params.

        When your site load is variable, it would be nice to be able to add
        workers dynamically. Enabling Zerg mode you can allow zerg clients to attach
        to your already running server and help it in the work.

        * http://uwsgi-docs.readthedocs.io/en/latest/Zerg.html

        :param str|unicode socket: Unix socket to bind server to.

            Examples:
                * unix socket - ``/var/run/mutalisk``
                * Linux abstract namespace - ``@nydus``

        :param str|unicode|list[str|unicode] clients_socket_pool: This enables Zerg Pools.

            .. note:: Expects master process.

            Accepts sockets that will be mapped to Zerg socket.

            * http://uwsgi-docs.readthedocs.io/en/latest/Zerg.html#zerg-pools

        """
        if clients_socket_pool:
            self._set('zergpool', '%s:%s' % (socket, ','.join(listify(clients_socket_pool))), multi=True)

        else:
            self._set('zerg-server', socket)

        return self._section

    def set_zerg_client_params(self, server_sockets, use_fallback_socket=None):
        """Zerg mode. Zergs params.

        :param str|unicode|list[str|unicode] server_sockets: Attaches zerg to a zerg server.

        :param bool use_fallback_socket: Fallback to normal sockets if the zerg server is not available

        """
        self._set('zerg', server_sockets, multi=True)

        if use_fallback_socket is not None:
            self._set('zerg-fallback', use_fallback_socket, cast=bool)

            for socket in listify(server_sockets):
                self._section.networking.register_socket(socket, type=self._section.networking.socket_types.DEFAULT)


        return self._section


class Cheapening(OptionsGroup):
    """uWSGI provides the ability to dynamically scale
    the number of running workers (adaptive process spawning) via pluggable algorithms.

    .. note:: This uses master process.

    """

    class algorithms(object):
        """Algorithms available to use with ``cheaper_algorithm``."""

        SPARE = 'spare'
        """The **default** algorithm. 
        
        If all workers are busy for ``cheaper_overload`` seconds then uWSGI will spawn new workers. 
        When the load is gone it will begin stopping processes one at a time.
        
        * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#spare-cheaper-algorithm
        
        """

        SPARE2 = 'spare2'
        """This algorithm is similar to ``spare``, but suitable for large scale 
        by increase workers faster and decrease workers slower.
        
        * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#spare2-cheaper-algorithm
        
        """

        BACKLOG = 'backlog'
        """If the socket's listen queue has more than ``cheaper_overload`` requests 
        waiting to be processed, uWSGI will spawn new workers. 
        
        If the backlog is lower it will begin killing processes one at a time.
        
        * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#backlog-cheaper-algorithm
        
        .. note: Only available on Linux and only on TCP sockets (not UNIX domain sockets).
        
        """

        BUSYNESS = 'busyness'
        """Algorithm adds or removes workers based on average utilization 
        for a given time period. It's goal is to keep more workers than 
        the minimum needed available at any given time, so the app will always 
        have capacity for new requests.
        
        * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#busyness-cheaper-algorithm
        
        .. note:: Requires ``cheaper_busyness`` plugin.
        
        """

        MANUAL = 'manual'
        """Algorithm allows to adjust number of workers using Master FIFO commands.
        
        * http://uwsgi.readthedocs.io/en/latest/MasterFIFO.html#available-commands
        
        """

    def set_basic_params(
            self, spawn_on_request=None,
            cheaper_algorithm=None, workers_min=None, workers_startup=None, workers_step=None):
        """

        :param bool spawn_on_request: Spawn workers only after the first request.

        :param cheaper_algorithm: The algorithm to be used used for adaptive process spawning. Default: ``spare``.
            See ``.algorithms``.

        :param int workers_min: Minimal workers count. Enables cheaper mode (adaptive process spawning).

            .. note:: Must be lower than max workers count.

        :param int workers_startup: The number of workers to be started when starting the application.
            After the app is started the algorithm can stop or start workers if needed.

        :param int workers_step: Number of additional processes to spawn at a time if they are needed,

        """
        self._set('cheap', spawn_on_request, cast=bool)
        self._set('cheaper-algo', cheaper_algorithm)
        self._set('cheaper', workers_min)
        self._set('cheaper-initial', workers_startup)
        self._set('cheaper-step', workers_step)

        return self._section

    def set_algo_spare_params(self, check_interval_overload=None):
        """Sets ``spare`` cheaper algorithm params.

        * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#spare-cheaper-algorithm

        :param int check_interval_overload: Interval (sec) to for worker overload.

        """
        self._set('cheaper-overload', check_interval_overload)

        return self._section

    def set_algo_spare2_params(self, check_interval_idle=None):
        """Sets ``spare2`` cheaper algorithm params.

        * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#spare2-cheaper-algorithm

        :param int check_interval_idle: Decrease workers after specified idle. Default: 10.

        """
        self._set('cheaper-idle', check_interval_idle)

        return self._section

    def set_algo_backlog_params(self, check_num_overload=None):
        """Sets ``backlog`` cheaper algorithm params.

        * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#backlog-cheaper-algorithm

        :param int check_num_overload: Number of backlog items in queue.

        """
        self._set('cheaper-overload', check_num_overload)

        return self._section

    def set_algo_busyness_params(
            self, check_interval_busy=None,
            busy_max=None, busy_min=None,
            idle_cycles_max=None, idle_cycles_penalty=None,
            verbose=None):
        """Sets busyness algorithm related params.

        :param int check_interval_busy: Interval (sec) to check worker busyness.

        :param int busy_max: Maximum busyness (percents). Every time the calculated busyness
            is higher than this value, uWSGI will spawn new workers. Default: 50.

        :param int busy_min: Minimum busyness (percents). If busyness is below this value,
            the app is considered in an "idle cycle" and uWSGI will start counting them.
            Once we reach needed number of idle cycles uWSGI will kill one worker. Default: 25.

        :param int idle_cycles_max: This option tells uWSGI how many idle cycles are allowed
            before stopping a worker.

        :param int idle_cycles_penalty: Number of idle cycles to add to ``idle_cycles_max``
            in case worker spawned too early. Default is 1.

        :param bool verbose: Enables debug logs for this algo.

        """
        set_ = partial(self._set, plugin='cheaper_busyness')

        set_('cheaper-overload', check_interval_busy)
        set_('cheaper-busyness-max', busy_max)
        set_('cheaper-busyness-min', busy_min)
        set_('cheaper-busyness-multiplier', idle_cycles_max)
        set_('cheaper-busyness-penalty', idle_cycles_penalty)
        set_('cheaper-busyness-verbose', verbose, cast=bool)

        return self._section

    def set_algo_busyness_emergency_params(
            self, workers_step=None, idle_cycles_max=None, backlog_size=None, backlog_nonzero_delay=None):
        """Sets busyness algorithm emergency workers related params.

        Emergency workers could be spawned depending upon uWSGI backlog state.

        .. note:: These options are Linux only.

        :param int workers_step: Number of emergency workers to spawn. Default: 1.

        :param int idle_cycles_max: Idle cycles to reach before stopping an emergency worker. Default: 3.

        :param int backlog_size: Backlog max queue size to spawn an emergency worker. Default: 33.

        :param int backlog_nonzero_delay: If the request listen queue is > 0 for more than given amount of seconds
            new emergency workers will be spawned. Default: 60.

        """
        set_ = partial(self._set, plugin='cheaper_busyness')

        set_('cheaper-busyness-backlog-step', workers_step)
        set_('cheaper-busyness-backlog-multiplier', idle_cycles_max)
        set_('cheaper-busyness-backlog-alert', backlog_size)
        set_('cheaper-busyness-backlog-nonzero', backlog_nonzero_delay)

        return self._section

    def set_memory_limits(self, rss_soft=None, rss_hard=None):
        """Sets worker memory limits for cheapening.

        :param int rss_soft: Don't spawn new workers if total resident memory usage
            of all workers is higher than this limit in bytes.

        :param int rss_hard: Try to stop workers if total workers resident memory usage
            is higher that thi limit in bytes.

        """
        self._set('cheaper-rss-limit-soft', rss_soft)
        self._set('cheaper-rss-limit-hard', rss_hard)

        return self._section

    def print_alorithms(self):
        """Print out enabled cheaper algorithms."""

        self._set('cheaper-algo-list', True, cast=bool)

        return self._section



import sys

from ..base import OptionsGroup
from ..exceptions import ConfigurationError


AUTO = (1,)


class Python(OptionsGroup):
    """Python plugin options.

    .. note:: By default the plugin does not initialize the GIL.
        This means your app-generated threads will not run.
        If you need threads, remember to enable them with ``enable_threads``.

    """

    plugin = True

    def set_basic_params(
            self, version=AUTO, python_home=None, enable_threads=None, search_path=None,
            python_binary=None, tracebacker_path=None, plugin_dir=None, **kwargs):
        """

        :param str|unicode|int version: Python version plugin supports.

            Example:
                * 3 - version 3
                * <empty> - version 2
                * <default> - version deduced by uwsgiconf

        :param str|unicode python_home: Set python executable directory - PYTHONHOME/virtualenv.

        :param str|unicode search_path: Add directory (or an .egg or a glob) to the Python search path.

            .. note:: This can be specified up to 64 times.

        :param str|unicode python_binary: Set python program name.

        :param str|unicode tracebacker_path: Enable the uWSGI Python tracebacker.
            http://uwsgi-docs.readthedocs.io/en/latest/Tracebacker.html

        :param str|unicode plugin_dir: Directory to search for plugin.

        :param bool enable_threads: Enable threads in the embedded languages.
            This will allow to spawn threads in your app.

            .. warning:: Threads will simply *not work* if this option is not enabled.
                         There will likely be no error, just no execution of your thread code.

        """
        self._set_name(version)

        self._section.workers.set_thread_params(enable=enable_threads)
        self._set('py-tracebacker', tracebacker_path)
        self._set('py-program-name', python_binary)
        self._set('pyhome', python_home)
        self._set('pythonpath', search_path, multi=True)

        self._section.set_plugins_params(search_dirs=plugin_dir)

        return self._section

    def _set_name(self, version=AUTO):
        """Returns plugin name."""

        name = 'python'

        if version:
            if version is AUTO:
                version = sys.version_info[0]

                if version == 2:
                    version = ''

            name = '%s%s' % (name, version)

        self.name = name

    def _get_name(self, *args, **kwargs):
        name = self.name

        if not name:
            self._set_name()

        return self.name

    def set_app_args(self, *args):
        """Sets ``sys.argv`` for python apps.

        Examples:
            * pyargv="one two three" will set ``sys.argv`` to ``('one', 'two', 'three')``.

        :param args:
        """
        if args:
            self._set('pyargv', ' '.join(args))

        return self._section

    def set_wsgi_params(self, module=None, callable_name=None):
        """Set wsgi related parameters.

        :param str|unicode module:
            * load .wsgi file as the Python application
            * load a WSGI module as the application.

            .. note:: The module (sans ``.py``) must be importable, ie. be in ``PYTHONPATH``.

            Examples:
                * mypackage.my_wsgi_module -- read from `application` attr of mypackage/my_wsgi_module.py
                * mypackage.my_wsgi_module:my_app -- read from `my_app` attr of mypackage/my_wsgi_module.py

        :param str|unicode callable_name: Set WSGI callable name. Default: application.

        """
        module = module or ''

        if '/' in module:
            self._set('wsgi-file', module)

        else:
            self._set('wsgi', module)

        self._set('callable', callable_name)

        return self._section

    def eval_wsgi_entrypoint(self, code):
        """Evaluates Python code as WSGI entry point.

        :param str|unicode code:
        """
        self._set('eval', code)

        return self._section

    def set_autoreload_params(self, scan_interval=None, ignore_modules=None):
        """Sets autoreload related parameters.

        :param int scan_interval: Seconds. Monitor Python modules' modification times to trigger reload.

            .. warning:: Use only in development.

        :param list|st|unicode ignore_modules: Ignore the specified module during auto-reload scan.

        """
        self._set('py-auto-reload', scan_interval)
        self._set('py-auto-reload-ignore', ignore_modules, multi=True)

        return self._section

    def register_module_alias(self, alias, module_path, after_init=False):
        """Adds an alias for a module.

        http://uwsgi-docs.readthedocs.io/en/latest/PythonModuleAlias.html

        :param str|unicode alias:
        :param str|unicode module_path:
        :param bool after_init: add a python module alias after uwsgi module initialization
        """
        command = 'post-pymodule-alias' if after_init else 'pymodule-alias'
        self._set(command, '%s=%s' % (alias, module_path), multi=True)

        return self._section

    def import_module(self, modules, shared=False, into_spooler=False):
        """Imports a python module.

        :param list|str|unicode modules:

        :param bool shared: Import a python module in all of the processes.

        :param bool into_spooler: Import a python module in the spooler.
            http://uwsgi-docs.readthedocs.io/en/latest/Spooler.html

        """
        if all((shared, into_spooler)):
            raise ConfigurationError('Unable to set both `shared` and `into_spooler` flags')

        if into_spooler:
            command = 'spooler-python-import'
        else:
            command = 'shared-python-import' if shared else 'python-import'

        self._set(command, modules, multi=True)

        return self._section

    def run_module(self, module):
        """Runs a Python script in the uWSGI environment.

        :param str|unicode module:

        """
        self._set('pyrun', module)

        return self._section



from uwsgiconf import Section


def test_workers_basics(assert_lines):

    assert_lines([
        'workers = 3',
    ], Section().workers.set_basic_params(count=3))

    assert_lines([
        'threads = 2',
    ], Section().workers.set_thread_params(count=2))

    assert_lines([
        'min-worker-lifetime = 10',
    ], Section().workers.set_reload_params(min_lifetime=10))

    assert_lines([
        'reload-on-exception = true',
    ], Section().workers.set_reload_on_exception_params(do_reload=True))

    assert_lines([
        'harakiri = 20',
    ], Section().workers.set_harakiri_params(timeout=20))

    assert_lines([
        'worker-exec = date',
    ], Section().workers.run_command_as_worker('date'))


def test_mules(assert_lines):

    assert_lines([
        'mule-harakiri = 3',
    ], Section().workers.set_mules_params(harakiri_timeout=3))

    section = Section()
    assert_lines([
        'farm = first:1,2',
        'farm = second:3,4,5',
        'mules = 5',

    ], section.workers.set_mules_params(farms=[
        section.workers.mule_farm('first', 2),
        section.workers.mule_farm('second', 3),
    ]))

    section = Section()
    assert_lines([
        'farm = first:1',
        'farm = second:2,3,4',
        'mules = 4',

    ], section.workers.set_mules_params(farms=[
        section.workers.mule_farm('first', [1]),
        section.workers.mule_farm('second', [2, 3, 4]),
    ]))

    assert_lines([
        'mule = mule1.py',
        'mule = mule2.py',

    ], Section().workers.set_mules_params(mules=[
        'mule1.py',
        'mule2.py',
    ]))


def test_zergs(assert_lines):

    assert_lines([
        'zerg-server = /here',
    ], Section().workers.set_zerg_server_params('/here'))

    assert_lines([
        'zergpool = /here:127.0.0.1:3031,127.0.0.1:3032',
    ], Section().workers.set_zerg_server_params(
        '/here', clients_socket_pool=['127.0.0.1:3031', '127.0.0.1:3032']))

    assert_lines([
        'zerg = /here',
        'zerg = /there',
        'zerg-fallback = true',
        'socket = /here',
        'socket = /there',

    ], Section().workers.set_zerg_client_params(
        ['/here', '/there'], use_fallback_socket=True
    ))


def test_cheapening(assert_lines):

    assert_lines([
        'cheaper-algo = manual',
    ], Section().cheapening.set_basic_params(cheaper_algorithm=Section.cheapening.algorithms.MANUAL))

    assert_lines([
        'cheaper-overload = 20',
    ], Section().cheapening.set_algo_spare_params(check_interval_overload=20))

    assert_lines([
        'cheaper-idle = 10',
    ], Section().cheapening.set_algo_spare2_params(check_interval_idle=10))

    assert_lines([
        'cheaper-overload = 30',
    ], Section().cheapening.set_algo_backlog_params(check_num_overload=30))

    assert_lines([
        'plugin = cheaper_busyness',
        'cheaper-busyness-max = 25',
    ], Section().cheapening.set_algo_busyness_params(busy_max=25))

    assert_lines([
        'plugin = cheaper_busyness',
        'cheaper-busyness-backlog-step = 3',
    ], Section().cheapening.set_algo_busyness_emergency_params(workers_step=3))

    assert_lines([
        'cheaper-rss-limit-soft = 1024',
    ], Section().cheapening.set_memory_limits(rss_soft=1024))

    assert_lines([
        'cheaper-algo-list = true',
    ], Section().cheapening.print_alorithms())



from ..base import OptionsGroup


class Locks(OptionsGroup):
    """Locks.

    * http://uwsgi.readthedocs.io/en/latest/Locks.html

    """

    def set_basic_params(self, count=None, thunder_lock=None, lock_engine=None):
        """
        :param int count: Create the specified number of shared locks.

        :param bool thunder_lock: Serialize accept() usage (if possible)
            Could improve performance on Linux with robust pthread mutexes.

            http://uwsgi.readthedocs.io/en/latest/articles/SerializingAccept.html

        :param str|unicode lock_engine: Set the lock engine.

            Example:
                - ipcsem

        """
        self._set('thunder-lock', thunder_lock, cast=bool)
        self._set('lock-engine', lock_engine)
        self._set('locks', count)

        return self._section

    def set_ipcsem_params(self, ftok=None, persistent=None):
        """Sets ipcsem lock engine params.

        :param str|unicode ftok: Set the ipcsem key via ftok() for avoiding duplicates.

        :param bool persistent: Do not remove ipcsem's on shutdown.

        """
        self._set('ftok', ftok)
        self._set('persistent-ipcsem', persistent, cast=bool)

        return self._section

    def lock_file(self, fpath, after_setup=False, wait=False):
        """Locks the specified file.

        :param str|unicode fpath: File path.

        :param bool after_setup:
            True  - after logging/daemon setup
            False - before starting

        :param bool wait:
            True  - wait if locked
            False - exit if locked

        """
        command = 'flock-wait' if wait else 'flock'

        if after_setup:
            command = '%s2' % command

        self._set(command, fpath)

        return self._section
